from __future__ import annotations

import logging

import pytest

from sibilant import PhoneState
from sibilant.voip import VoIPPhone


_logger = logging.getLogger(__name__)


@pytest.mark.needs_test_server
class TestVoIPPhoneReal:
    def test_start_stop(self, test_server_kwargs):
        phone = VoIPPhone(**test_server_kwargs)

        phone.start()
        assert phone.registered, "Phone should be registered with the server"
        assert phone.state == PhoneState.READY, "Phone state should be READY"

        phone.stop()
        assert not phone.registered, "Phone should be unregistered with the server"
        assert phone.state == PhoneState.INACTIVE, "Phone state should be INACTIVE"

    def test_start_stop_context(self, test_server_kwargs):
        with VoIPPhone(**test_server_kwargs) as phone:
            assert phone.registered, "Phone should be registered with the server"
            assert phone.state == PhoneState.READY, "Phone state should be READY"

        assert not phone.registered, "Phone should be unregistered with the server"
        assert phone.state == PhoneState.INACTIVE, "Phone state should be INACTIVE"



"""SIP headers classes."""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import field as dataclass_field, fields as dataclass_fields
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    TypeVar,
    cast,
)

from typing_extensions import Self, override

from sibilant.exceptions import SIPParseError
from sibilant.helpers import (
    DEFAULT,
    CaseInsensitiveDict,
    DefaultType,
    IntValueMixin,
    ListValueMixin,
    ParseableSerializable,
    Registry,
    StrValueMixin,
    SupportsStr,
    slots_dataclass,
)
from sibilant.structures import SIPAddress


if TYPE_CHECKING:
    from collections.abc import Mapping

    from .messages import SIPMethod


__all__ = [
    "AllowHeader",
    "AuthorizationHeader",
    "CSeqHeader",
    "CallIDHeader",
    "Contact",
    "ContactHeader",
    "ContentLengthHeader",
    "ContentTypeHeader",
    "ExpiresHeader",
    "FromHeader",
    "FromToHeader",
    "Header",
    "Headers",
    "IntHeader",
    "ListHeader",
    "MaxForwardsHeader",
    "MultipleValuesHeader",
    "PAssertedIdentityHeader",
    "PPreferredIdentityHeader",
    "ProxyAuthenticateHeader",
    "ProxyAuthorizationHeader",
    "RecordRouteHeader",
    "RouteHeader",
    "StrHeader",
    "SupportedHeader",
    "ToHeader",
    "UnknownHeader",
    "UserAgentHeader",
    "ViaEntry",
    "ViaHeader",
    "WWWAuthenticateHeader",
]


_H = TypeVar("_H", bound="Header")


class Header(
    Registry[str | DefaultType, "Header"],
    ABC,
    registry=True,
    registry_attr="_name",
):
    """Abstract base dataclass for SIP headers."""

    _name: ClassVar[str | DefaultType]

    @property
    def name(self) -> str:
        """The name of the header."""
        if self._name is DEFAULT:
            raise SyntaxError(
                f"Class {self.__class__} must override name() property when using _name = DEFAULT"
            )
        assert isinstance(self._name, str)
        return self._name

    @property
    def raw_value(self) -> str:
        """The raw value of the header."""
        return self.serialize()

    @classmethod
    def parse(
        cls, header: str, values: str | list[str], previous_headers: Headers
    ) -> Self:
        """
        Parse a raw header into a header object, picking the correct class.

        :param header: the header name
        :param values: the raw values of the header
        :param previous_headers: the previous headers that have been already parsed
        :return: the new header object.
        """
        known_header: bool = header in cls.__registry__
        if not known_header and DEFAULT not in cls.__registry__:
            raise TypeError(
                f"Unknown header {header}, and no default header class is defined"
            )
        header_cls: type[Header] = cls.__registry_get_class_for__(
            header if known_header else DEFAULT
        )
        if isinstance(values, str):
            values = [values]
        assert values, "At least one value must be present"
        if len(values) > 1:
            if not issubclass(header_cls, MultipleValuesHeader):
                raise SIPParseError(
                    f"Multiple values for header {header}, but header is not a MultipleValuesHeader"
                )
            value = ListHeader._separator.join(values)  # noqa: SLF001
        else:
            value = values[0]
        return cast("Self", header_cls.from_raw_value(header, value, previous_headers))

    @classmethod
    @abstractmethod
    def from_raw_value(cls, header: str, value: str, previous_headers: Headers) -> Self:
        """
        Parse the header value from a string.

        :param header: the header name
        :param value: The raw header value string.
        :param previous_headers: The previous headers in the message.
        :return: The parsed header.
        """

    @abstractmethod
    def serialize(self) -> str:
        """Serialize the header value to a string."""

    def __str__(self) -> str:
        """Serialize the entire header to a string."""
        return f"{self.name}: {self.serialize()}"


@slots_dataclass
class StrHeader(StrValueMixin, Header, ABC):
    """Abstract base dataclass for headers with a single string value."""

    @classmethod
    def from_raw_value(cls, header: str, value: str, previous_headers: Headers) -> Self:  # noqa: D102
        return cls(**cls.parse_raw_value(value))


@slots_dataclass
class UnknownHeader(StrHeader):
    """Catch-all dataclass for unsupported SIP headers."""

    _name = DEFAULT

    header: str

    @property
    @override
    def name(self) -> str:
        return self.header

    @classmethod
    @override
    def from_raw_value(cls, header: str, value: str, previous_headers: Headers) -> Self:
        return cls(header=header, value=value)


@slots_dataclass
class IntHeader(IntValueMixin, Header, ABC):
    """Abstract base dataclass for headers with a single integer value."""

    @classmethod
    def from_raw_value(cls, header: str, value: str, previous_headers: Headers) -> Self:  # noqa: D102
        return cls(**cls.parse_raw_value(value))


_ST = TypeVar("_ST", bound=SupportsStr | ParseableSerializable)


class ListHeader(ListValueMixin[_ST], Header, ABC):
    """Abstract base dataclass for headers with a list of values."""

    @classmethod
    def from_raw_value(cls, header: str, value: str, previous_headers: Headers) -> Self:  # noqa: D102
        return cls(**cls.parse_raw_value(value))


class MultipleValuesHeader(ListHeader[_ST], ABC):
    """Abstract base dataclass for headers which can appear multiple times."""

    _prefers_separate_lines: ClassVar[bool] = False

    @override
    def __str__(self) -> str:
        if self._prefers_separate_lines:
            return "\r\n".join(f"{self.name}: {v}" for v in self._serialized_values())
        return f"{self.name}: {self.serialize()}"


@slots_dataclass
class ViaEntry(ParseableSerializable):
    """A single Via entry as part of a Via header, as described in :rfc:`3261#section-20.42`."""

    method: str
    address: str
    port: int | None
    branch: str | None = None
    maddr: str | None = None
    received: str | None = None
    ttl: int | None = None
    extension: dict[str, str | None] = dataclass_field(default_factory=dict)

    _order: tuple[str, ...] = ()

    @property
    def rport(self) -> bool | int | None:
        """The rport parameter, if present."""
        if self.extension and "rport" in self.extension:
            return int(self.extension["rport"]) if self.extension["rport"] else True
        return None

    @rport.setter
    def rport(self, value: bool | int | None) -> None:
        if not value:
            self.extension.pop("rport", None)
        elif isinstance(value, bool):
            self.extension["rport"] = None
        else:
            assert isinstance(value, int)
            self.extension["rport"] = str(value)

    @classmethod
    def parse(cls, value: str) -> Self:  # noqa: D102
        method, address, *params = re.split(r"\s+|\s*;\s*", value.strip())

        ip, port_str = address.split(":") if ":" in address else (address, None)
        port = int(port_str) if port_str else None

        order = []
        parsed_params: dict[str, str | int | None] = {}
        extension_params: dict[str, str | None] = {}
        for param in params:
            param_name: str
            param_value: str | int | None
            param_name, param_value = (
                param.split("=", maxsplit=1) if "=" in param else (param, None)
            )
            if param_name == "ttl":
                param_value = int(param_value) if param_value is not None else None
            assert isinstance(param_value, str) or param_value is None
            if param_name in {"branch", "maddr", "received", "ttl"}:
                parsed_params[param_name] = param_value
            else:
                extension_params[param_name] = param_value
            order.append(param_name)

        return cls(
            method=method,
            address=ip,
            port=port,
            **parsed_params,  # type: ignore[arg-type]
            extension=extension_params,
            _order=tuple(order),
        )

    def serialize(self) -> str:  # noqa: D102
        host = f"{self.address}:{self.port}" if self.port else self.address
        params_dict = {
            param_name: param_value
            for param_name in ("branch", "maddr", "received", "ttl")
            if (param_value := getattr(self, param_name)) is not None
        }
        params_dict.update(self.extension or {})
        sorted_params = {**dict.fromkeys(self._order), **params_dict}
        params = [
            f";{param_name}={param_value}"
            if param_value is not None
            else f";{param_name}"
            for param_name, param_value in (sorted_params.items())
        ]
        return f"{self.method} {host}{''.join(params)}"


@slots_dataclass
class ViaHeader(MultipleValuesHeader[ViaEntry]):
    """Via header, as described in :rfc:`3261#section-20.42`."""

    _name = "Via"
    _prefers_separate_lines = True
    _values_type = ViaEntry

    @property
    def first(self) -> ViaEntry:
        """The first via entry in the header."""
        return self.values[0]

    def __post_init__(self) -> None:
        if not self.values:
            raise SIPParseError("Via header must have at least one entry")


@slots_dataclass
class FromToHeader(Header, ABC):
    """Abstract base dataclass for From and To headers."""

    address: SIPAddress
    raw_address: str | None = None
    tag: str | None = None

    def __post_init__(self) -> None:
        if self.raw_address is None:
            self.raw_address = str(self.address)

    @classmethod
    def from_raw_value(cls, header: str, value: str, previous_headers: Headers) -> Self:  # noqa: D102
        raw_address, *taginfo = re.split(r"\s*;\s*tag=", value.strip(), maxsplit=1)
        tag = taginfo[0] if taginfo else None
        address = SIPAddress.parse(raw_address)
        return cls(address=address, raw_address=raw_address, tag=tag)

    def serialize(self) -> str:  # noqa: D102
        assert self.raw_address is not None
        if self.tag:
            return f"{self.raw_address};tag={self.tag}"
        return self.raw_address


@slots_dataclass
class FromHeader(FromToHeader):
    """From header, as described in :rfc:`3261#section-20.20`."""

    _name = "From"


@slots_dataclass
class ToHeader(FromToHeader):
    """To header, as described in :rfc:`3261#section-20.39`."""

    _name = "To"


@slots_dataclass
class Contact(ParseableSerializable):
    """A single contact as part of a contact header, as described in :rfc:`3261#section-20.10`."""

    address: SIPAddress
    params: dict[str, str | None] = dataclass_field(default_factory=dict)

    @property
    def q(self) -> float | None:
        """The q parameter, if present."""
        if self.params and isinstance(
            (q_str := self.params.get("q")), (str, int, float)
        ):
            return float(q_str)
        return None

    @q.setter
    def q(self, value: float | None) -> None:
        if self.params is None:
            self.params = {}
        if value is None:
            self.params.pop("q", None)
        else:
            self.params["q"] = str(value)

    @property
    def expires(self) -> int | None:
        """The expires parameter, if present."""
        if self.params and isinstance(
            (expires_str := self.params.get("expires")), (str, int, float)
        ):
            return int(expires_str)
        return None

    @expires.setter
    def expires(self, value: int | None) -> None:
        if self.params is None:
            self.params = {}
        if value is None:
            self.params.pop("expires", None)
        else:
            self.params["expires"] = str(value)

    @classmethod
    def parse(cls, value: str) -> Self:  # noqa: D102
        params: dict[str, str | None] = {}

        force_brackets: bool = "<" in value and ">" in value
        if ";" in value:  # there are parameters, address will be in <...> form.
            address_raw, *params_raw = re.split(r"(?<=>)\s*;\s*", value, maxsplit=1)
            if params_raw:
                for param in re.split(r"\s*;\s*", params_raw[0]):
                    param_name, param_value = param.split("=", maxsplit=1)
                    param_name = param_name.strip()
                    param_value = param_value.strip()
                    params[param_name] = param_value
            force_brackets = True
        else:
            address_raw = value

        address = SIPAddress.parse(address_raw.strip(), force_brackets=force_brackets)
        return cls(address=address, params=params)

    def serialize(self) -> str:  # noqa: D102
        params = [f";{name}={value}" for name, value in self.params.items()]
        return f"{self.address}{''.join(params)}"


@slots_dataclass
class ContactHeader(MultipleValuesHeader[Contact]):
    """
    Contact header, as described in :rfc:`3261#section-20.10`.

    Multiple contacts might be present in a single header, separated by commas.
    """

    _name = "Contact"
    _values_type = Contact

    # TODO: * contact (see RFC?)

    @property
    def contact(self) -> Contact:
        """The first contact in the header."""
        return self.values[0]


@slots_dataclass
class RouteHeader(ContactHeader):
    """Route header, as described in :rfc:`3261#section-20.34`."""

    _name = "Route"


@slots_dataclass
class RecordRouteHeader(RouteHeader):
    """Record-Route header, as described in :rfc:`3261#section-20.30`."""

    _name = "Record-Route"


@slots_dataclass
class CallIDHeader(StrHeader):
    """Call-ID header, as described in :rfc:`3261#section-20.8`."""

    _name = "Call-ID"


@slots_dataclass
class CSeqHeader(Header):
    """CSeq header, as described in :rfc:`3261#section-20.16`."""

    _name = "CSeq"

    sequence: int
    method: SIPMethod

    @classmethod
    def from_raw_value(cls, header: str, value: str, previous_headers: Headers) -> Self:  # noqa: D102
        from .messages import SIPMethod  # noqa: PLC0415

        sequence, method_raw = value.split(maxsplit=1)
        method = SIPMethod.match_value(method_raw)
        return cls(sequence=int(sequence), method=method)

    def serialize(self) -> str:  # noqa: D102
        return f"{self.sequence} {self.method.name}"


@slots_dataclass
class AllowHeader(MultipleValuesHeader[str]):
    """Allow header, as described in :rfc:`3261#section-20.5`."""

    _name = "Allow"
    _values_type = str


@slots_dataclass
class AcceptHeader(MultipleValuesHeader[str]):
    """Accept header, as described in :rfc:`3261#section-20.1`."""

    _name = "Accept"
    _values_type = str


@slots_dataclass
class SupportedHeader(MultipleValuesHeader[str]):
    """Supported header, as described in :rfc:`3261#section-20.37`."""

    _name = "Supported"
    _values_type = str


@slots_dataclass
class ExpiresHeader(IntHeader):
    """Expires header, as described in :rfc:`3261#section-20.19`."""

    _name = "Expires"


@slots_dataclass
class ContentTypeHeader(StrHeader):
    """Content-Type header, as described in :rfc:`3261#section-20.15`."""

    _name = "Content-Type"


@slots_dataclass
class ContentLengthHeader(IntHeader):
    """Content-Length header, as described in :rfc:`3261#section-20.14`."""

    _name = "Content-Length"


@slots_dataclass
class MaxForwardsHeader(IntHeader):
    """Max-Forwards header, as described in :rfc:`3261#section-20.22`."""

    _name = "Max-Forwards"


@slots_dataclass
class UserAgentHeader(StrHeader):
    """User-Agent header, as described in :rfc:`3261#section-20.41`."""

    _name = "User-Agent"


@slots_dataclass
class AuthorizationHeader(Header):
    """Authorization header, as described in :rfc:`3261#section-20.7`."""

    _name = "Authorization"

    username: str | None = None
    realm: str | None = None
    nonce: str | None = None
    uri: str | None = None
    algorithm: str | None = None
    qop: str | None = None
    nc: str | None = None
    cnonce: str | None = None
    response: str | None = None
    opaque: str | None = None
    stale: bool | None = None
    auth_params: dict[str, str] | None = None

    _order: tuple[str, ...] = ()

    _no_quote_params: ClassVar[set[str]] = {"algorithm", "stale", "qop", "nc"}

    @classmethod
    def from_raw_value(cls, header: str, value: str, previous_headers: Headers) -> Self:  # noqa: D102
        scheme, params_str = value.split(maxsplit=1)
        if scheme.lower() != "digest":
            raise SIPParseError(f"Unsupported authorization scheme: {scheme}")
        # split by commas, remove whitespaces, split by equal sign, remove quotes
        # FIXME: quoted values might contain commas!
        params = {
            key: value.strip('"')
            for key, value in (
                param.strip().split("=", maxsplit=1)
                for param in params_str.strip().split(",")
            )
        }
        # TODO: qop can be a comma-separated list of values and might be quoted
        known_param_names = {f.name for f in dataclass_fields(cls)} - {"auth_params"}
        known_params: dict[str, Any] = {
            name: value for name, value in params.items() if name in known_param_names
        }
        if "stale" in known_params:
            known_params["stale"] = known_params["stale"].lower() == "true"
        auth_params = {
            name: value for name, value in params.items() if name not in known_params
        }
        return cls(
            **known_params,
            auth_params=auth_params or None,
            _order=tuple(params),  # type: ignore[arg-type]
        )

    def serialize(self) -> str:  # noqa: D102
        params_dict: dict[str, str] = {}
        for field in dataclass_fields(self):
            if field.name in {"auth_params", "_order"}:
                continue
            value = getattr(self, field.name)
            if value is not None:
                str_value = str(value)
                if isinstance(value, bool):
                    str_value = str_value.upper()
                params_dict[field.name] = str_value
        if self.auth_params:
            # noinspection PyTypeChecker
            params_dict.update(self.auth_params)
        sorted_params = {**dict.fromkeys(self._order), **params_dict}
        params = [
            (name, (f'"{value}"' if name not in self._no_quote_params else str(value)))
            for name, value in sorted_params.items()
        ]
        return f"Digest {', '.join(f'{name}={value}' for name, value in params)}"


@slots_dataclass
class WWWAuthenticateHeader(AuthorizationHeader):
    """WWW-Authenticate header, as described in :rfc:`3261#section-20.44`."""

    _name = "WWW-Authenticate"


@slots_dataclass
class ProxyAuthorizationHeader(AuthorizationHeader):
    """Proxy-Authorization header, as described in :rfc:`3261#section-20.28`."""

    _name = "Proxy-Authorization"


@slots_dataclass
class ProxyAuthenticateHeader(WWWAuthenticateHeader):
    """Proxy-Authenticate header, as described in :rfc:`3261#section-20.27`."""

    _name = "Proxy-Authenticate"


@slots_dataclass
class PAssertedIdentityHeader(MultipleValuesHeader[SIPAddress]):
    """P-Asserted-Identity header, as described in :rfc:`3325#section-9.1`."""

    _name = "P-Asserted-Identity"
    _values_type = SIPAddress

    def __post_init__(self) -> None:
        if not self.values:
            raise SIPParseError(
                f"{self._name} header if present must have at least one entry"
            )
        if len(self.values) > 2:
            raise SIPParseError(f"{self._name} header can have only up to two entries")


@slots_dataclass
class PPreferredIdentityHeader(MultipleValuesHeader[SIPAddress]):
    """P-Preferred-Identity header, as described in :rfc:`3325#section-9.2`."""

    _name = "P-Preferred-Identity"
    _values_type = SIPAddress

    def __post_init__(self) -> None:
        if not self.values:
            raise SIPParseError(
                f"{self._name} header if present must have at least one entry"
            )
        if len(self.values) > 2:
            raise SIPParseError(f"{self._name} header can have only up to two entries")


class Headers(CaseInsensitiveDict[_H]):
    """
    A case-insensitive dictionary of SIP headers, with some additional parsing
    and serialization methods to handle raw headers from/to SIP messages.

    The dictionary keys are the header names, and the values are the parsed header objects.

    :param headers: the headers to initialize the dictionary with.
    :param data: additional headers to initialize the dictionary with, as a mapping.
    :param kwargs: additional headers to initialize the dictionary with, as keyword arguments.
    """

    def __init__(
        self, *headers: _H, data: Mapping[str, _H] | None = None, **kwargs: _H
    ):
        if headers:
            data = dict(data) if data else {}
            for header in headers:
                data[header.name] = header

        super().__init__(data, **kwargs)

    @classmethod
    def parse(cls, raw_headers: bytes) -> Self:
        """
        Parse a raw SIP message into a dictionary of headers.

        :param raw_headers: the raw headers
        :return: the parsed headers.
        """
        headers = cls()

        header_lines = raw_headers.decode("utf-8").split("\r\n")
        headers_values = defaultdict(list)
        for line in header_lines:
            header, raw_value = line.split(": ", maxsplit=1)
            headers_values[header].append(raw_value)

        for header, raw_values in headers_values.items():
            headers[header] = cast("_H", Header.parse(header, raw_values, headers))

        return headers

    def serialize(self) -> bytes:
        """
        Serialize the headers to a raw SIP message.

        :return: the raw headers.
        """
        return str(self).encode("utf-8")

    def __str__(self) -> str:
        return "\r\n".join(str(header) for header in self.values())



"""SDP time section and fields definitions and implementations."""

from __future__ import annotations

import re
from abc import ABC

from typing_extensions import Self, override

from sibilant.exceptions import SDPParseError
from sibilant.helpers import slots_dataclass

from .common import SDPField, SDPSection


__all__ = [
    "SDPTime",
    "SDPTimeFields",
    "SDPTimeRepeat",
    "SDPTimeTime",
]


@slots_dataclass
class SDPTimeFields(SDPField, ABC, registry=True, registry_attr="_type"):
    """Base class for SDP time description fields."""


@slots_dataclass
class SDPTimeTime(SDPTimeFields):
    """
    SDP time field, defined in :rfc:`8866#section-5.9`.

    Spec::
        t=<start-time> <stop-time>
    """

    _type = "t"
    _description = "time the session is active"

    start_time: int
    stop_time: int

    @classmethod
    @override
    def from_raw_value(cls, field_type: str, raw_value: str) -> Self:
        start_time, stop_time = raw_value.split(" ")
        return cls(start_time=int(start_time), stop_time=int(stop_time))

    def serialize(self) -> str:  # noqa: D102
        return f"{self.start_time} {self.stop_time}"


@slots_dataclass
class SDPTimeRepeat(SDPTimeFields):
    """
    SDP time repeat field, defined in :rfc:`8866#section-5.10`.

    Spec::
        r=<repeat interval> <active duration> <offsets from start-time>
    """

    _type = "r"
    _description = "zero or more repeat times"

    interval: int
    duration: int
    offsets: list[int]

    @classmethod
    @override
    def from_raw_value(cls, field_type: str, raw_value: str) -> Self:
        # parse the raw value. N.B. offset could be strings to denote days, hours, minutes, seconds,
        # so they need to be converted to ints
        def parse_time(time_str: str) -> int:
            if isinstance(time_str, int):
                return time_str
            multipliers = {"d": 86400, "h": 3600, "m": 60, "s": 1}
            match = re.match(rf"(-?\d+)([{''.join(multipliers)}])", time_str)
            if not match:
                raise SDPParseError(
                    f'Invalid time string "{time_str}" in repeat field: {raw_value}'
                )
            time, unit = match.groups()
            return int(time) * multipliers[unit]

        interval, duration, *offsets = raw_value.split(" ")
        return cls(
            interval=parse_time(interval),
            duration=parse_time(duration),
            offsets=[parse_time(offset) for offset in offsets],
        )

    def serialize(self) -> str:  # noqa: D102
        return f"{self.interval} {self.duration} {' '.join(str(offset) for offset in self.offsets)}"


@slots_dataclass
class SDPTime(SDPSection):
    """SDP section for time description fields, defined in :rfc:`8866#section-5.9`."""

    _fields_base = SDPTimeFields
    _start_field = SDPTimeTime

    time: SDPTimeTime
    repeat: SDPTimeRepeat | None = None



from __future__ import annotations

import enum
import logging
import re
import socket
import threading
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass, replace as dataclass_replace
from pathlib import Path
from typing import TYPE_CHECKING, Generic, TypeVar

import dpkt
import pytest
from dpkt.utils import inet_to_str


if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import TracebackType


_logger = logging.getLogger(__name__)


def pytest_addoption(parser):
    parser.addoption("--test-server-address", help="Address of the test server")
    parser.addoption(
        "--test-server-username", help="Username to use for the test server"
    )
    parser.addoption(
        "--test-server-password", help="Password to use for the test server"
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "needs_test_server: mark test as needing a test server"
    )


def get_test_server_options(config):
    server_address = config.getoption("--test-server-address")
    server_kwargs = dict(server_host=server_address)
    if server_address and ":" in server_address:
        server_host, server_port = server_address.split(":")
        server_kwargs.update(server_host=server_host, server_port=int(server_port))
    phone_kwargs = dict(
        username=config.getoption("--test-server-username"),
        password=config.getoption("--test-server-password"),
        **server_kwargs,
    )
    if not any(phone_kwargs.values()):
        raise ValueError(
            "need --test-server command line options to run with a real server"
        )
    if not all(phone_kwargs.values()):
        raise ValueError(
            "need ALL these command line options to run with a real server: "
            "--test-server-address, --test-server-username, --test-server-password"
        )
    return phone_kwargs


def pytest_collection_modifyitems(config, items):
    enable_real_server_tests = False
    skip_real_server_tests_reason = "need --test-server command line options to run"
    try:
        get_test_server_options(config)
        enable_real_server_tests = True
    except ValueError as e:
        skip_real_server_tests_reason = str(e)

    skip_needs_test_server = pytest.mark.skip(reason=skip_real_server_tests_reason)
    for item in items:
        if "needs_test_server" in item.keywords and not enable_real_server_tests:
            item.add_marker(skip_needs_test_server)


@pytest.fixture(scope="session")
def test_server_kwargs(pytestconfig):
    return get_test_server_options(pytestconfig)


class Dest(enum.Enum):
    CLIENT = enum.auto()
    SERVER = enum.auto()


class PacketType(enum.Enum):
    SIP = enum.auto()
    RTP = enum.auto()


@dataclass(frozen=True)
class Packet:
    timestamp: float
    dest: Dest | None
    src_addr: tuple[str, int]
    dst_addr: tuple[str, int]
    type: PacketType
    data: bytes


@pytest.fixture(scope="session")
def voip_calls():  # noqa: PLR0914
    """
    Parse pcap files, and return an iterable of calls, which are iterables of packets.
    Separate client and server packets, and SIP and RTP packets.
    """
    calls = []

    pcaps_dir = Path(__file__).parent / "pcaps"
    for pcap_file in pcaps_dir.glob("*.pcap"):
        call_packets = []
        known_server_ips = set()
        known_client_ips = set()
        with open(pcap_file, "rb") as fp:
            pcap = dpkt.pcap.Reader(fp)

            for ts, buf in pcap:
                eth = dpkt.ethernet.Ethernet(buf)
                if eth.type != dpkt.ethernet.ETH_TYPE_IP:
                    continue

                ip = eth.data
                if ip.p not in {dpkt.ip.IP_PROTO_TCP, dpkt.ip.IP_PROTO_UDP}:
                    continue
                data, src, sport, dst, dport = (
                    ip.data.data,
                    ip.src,
                    ip.data.sport,
                    ip.dst,
                    ip.data.dport,
                )
                src_str = inet_to_str(src)
                aux_src_str = None
                dst_str = inet_to_str(dst)

                is_sip = b"SIP/2.0" in data
                # make sure it's not a RTCP packet, checking that packet type is not 200-204
                is_rtp = data.startswith(b"\x80") and not (200 <= data[1] <= 204)
                packet_type = (
                    PacketType.SIP if is_sip else (PacketType.RTP if is_rtp else None)
                )
                if packet_type is None:
                    continue

                def extract_aux_ip():
                    nonlocal aux_src_str
                    if match := re.search(
                        r"^c *= *IN +IP4 +\b(\d+\.\d+\.\d+\.\d+)\b",
                        data.decode(),  # noqa: B023
                        re.MULTILINE | re.IGNORECASE,
                    ):
                        aux_src_str = match.group(1)

                if is_sip:
                    first_line = data.split(b"\r\n")[0]
                    dest = None
                    if first_line.startswith(b"REGISTER"):
                        dest = Dest.SERVER
                    elif first_line.startswith(b"INVITE"):
                        if re.search(
                            rf"^Via:.*{src_str}",
                            data.decode(),
                            re.MULTILINE | re.IGNORECASE,
                        ):
                            dest = Dest.SERVER
                        else:
                            dest = Dest.CLIENT
                        extract_aux_ip()
                    elif first_line.startswith(b"SIP/2.0"):
                        if re.search(
                            r"^CSeq:.*REGISTER",
                            data.decode(),
                            re.MULTILINE | re.IGNORECASE,
                        ):
                            dest = Dest.CLIENT
                        elif re.search(
                            r"^CSeq:.*INVITE",
                            data.decode(),
                            re.MULTILINE | re.IGNORECASE,
                        ):
                            dest = Dest.SERVER
                            extract_aux_ip()

                    # server packets are sometimes recognized as client, so ignore that
                    if dest is not None:
                        if dest is Dest.SERVER and src_str not in known_server_ips:
                            known_client_ips.add(src_str)
                            if aux_src_str:
                                known_client_ips.add(aux_src_str)
                        elif dest is Dest.CLIENT:
                            known_client_ips.discard(src_str)
                            known_server_ips.add(src_str)
                            if aux_src_str:
                                known_server_ips.add(aux_src_str)

                call_packets.append(
                    Packet(
                        timestamp=ts,
                        dest=None,
                        type=packet_type,
                        src_addr=(src_str, sport),
                        dst_addr=(dst_str, dport),
                        data=data,
                    )
                )

        def categorize_packet(packet):
            if packet.src_addr[0] in known_server_ips:  # noqa: B023
                return Dest.CLIENT
            if packet.src_addr[0] in known_client_ips:  # noqa: B023
                return Dest.SERVER
            if packet.dst_addr[0] in known_server_ips:  # noqa: B023
                return Dest.SERVER
            if packet.dst_addr[0] in known_client_ips:  # noqa: B023
                return Dest.CLIENT
            return None

        # re-categorize client/server packets, having seen all packets
        call_packets = [
            dataclass_replace(p, dest=categorize_packet(p)) for p in call_packets
        ]

        uncategorized_packets = [p for p in call_packets if p.dest is None]
        if uncategorized_packets:
            raise ValueError(f"Uncategorized packets in {pcap_file}")

        if call_packets:
            calls.append(tuple(call_packets))

    return tuple(calls)


@pytest.fixture
def sip_packets(voip_calls):
    """Return a list of SIP packets from all the calls."""
    return [
        packet
        for call in voip_calls
        for packet in call
        if packet.type == PacketType.SIP
    ]


@pytest.fixture
def sip_requests(sip_packets):
    """Return a list of SIP requests from all the calls."""
    return [packet for packet in sip_packets if not packet.data.startswith(b"SIP/2.0")]


@pytest.fixture
def sip_responses(sip_packets):
    """Return a list of SIP responses from all the calls."""
    return [packet for packet in sip_packets if packet.data.startswith(b"SIP/2.0")]


@pytest.fixture
def sip_packets_from_client(sip_packets):
    """Return a list of SIP packets from the client."""
    return [packet for packet in sip_packets if packet.dest == Dest.SERVER]


@pytest.fixture
def sip_packets_from_server(sip_packets):
    """Return a list of SIP packets from the server."""
    return [packet for packet in sip_packets if packet.dest == Dest.CLIENT]


@pytest.fixture
def rtp_packets(voip_calls):
    """Return a list of RTP packets from all the calls."""
    return [
        packet
        for call in voip_calls
        for packet in call
        if packet.type == PacketType.RTP
    ]


@pytest.fixture
def rtp_packets_from_client(rtp_packets):
    """Return a list of RTP packets from the client."""
    return [packet for packet in rtp_packets if packet.dest == Dest.SERVER]


@pytest.fixture
def rtp_packets_from_server(rtp_packets):
    """Return a list of RTP packets from the server."""
    return [packet for packet in rtp_packets if packet.dest == Dest.CLIENT]


_PT = TypeVar("_PT")


class MockServer(ABC, Generic[_PT]):
    """Small mock UDP server, opens up a connection and sends/recvs data."""

    def __init__(
        self,
        packets_iterator: Iterator[_PT],
        server_address,
        client_address,
        send_delay=1e-6,
        recv_delay=1e-6,
        *,
        pre_bind=False,
        ready=True,
    ):
        self.packets_iterator: Iterator[_PT] = packets_iterator
        self.server_address = server_address
        self.client_address = client_address
        self.send_delay = send_delay
        self.recv_delay = recv_delay

        self.socket = None
        if pre_bind:
            self.socket = self._create_socket()
        self.send_thread = None
        self.recv_thread = None
        self.ready_event = threading.Event()
        if ready:
            self.ready_event.set()
        self.stop_event = threading.Event()
        self.error = None

    def _create_socket(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setblocking(False)
        sock.bind(self.server_address)
        return sock

    def start(self):
        if self.socket is None:
            self.socket = self._create_socket()

        self.send_thread = threading.Thread(
            target=self._run_send, daemon=True, name="MockServer._run_send"
        )
        self.recv_thread = threading.Thread(
            target=self._run_recv, daemon=True, name="MockServer._run_recv"
        )
        self.stop_event.clear()
        self.send_thread.start()
        self.recv_thread.start()

    @property
    def send_done(self):
        return self.send_thread is None or not self.send_thread.is_alive()

    @abstractmethod
    def send(self, packet: _PT):
        """Send a packet to the client."""

    def _run_send(self):
        self.ready_event.wait(timeout=30)
        while not self.stop_event.is_set():
            try:
                packet: _PT = next(self.packets_iterator)
                self.send(packet)
            except StopIteration:
                break
            except Exception as e:
                self.error = e
                self.stop_event.set()
                raise

            # FIXME: if we don't wait, we lose packets. Can we fix?
            #        does this have to do with the socket being non-blocking?
            time.sleep(self.send_delay)

    @abstractmethod
    def recv(self):
        """Receive a packet from the client."""

    def _run_recv(self):
        while not self.stop_event.is_set():
            try:
                self.recv()
            except Exception as e:
                self.error = e
                self.stop_event.set()
                raise

            time.sleep(self.recv_delay)

    def stop(self):
        self.stop_event.set()
        self.send_thread.join()
        self.recv_thread.join()
        self.socket.close()

    def __enter__(self):
        self.start()
        return self

    def __exit__(
        self,
        exctype: type[BaseException] | None,
        excinst: BaseException | None,
        exctb: TracebackType | None,
    ) -> None:
        self.stop()



from __future__ import annotations

import logging
import re
import time
import traceback
from collections import defaultdict, deque, namedtuple
from contextlib import contextmanager, nullcontext
from typing import TYPE_CHECKING, NamedTuple

import pytest

from sibilant import rtp
from sibilant.exceptions import SIPException
from sibilant.sip import (
    CallHandler,
    CallState,
    HashedAuth,
    Header,
    Headers,
    MultipleValuesHeader,
    SIPCall,
    SIPClient,
    SIPMessage,
    SIPMethod,
    SIPRegistration,
    SIPRequest,
    SIPResponse,
    compute_auth,
)
from sibilant.structures import SIPURI

from .conftest import Dest, MockServer


if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


_logger = logging.getLogger(__name__)


class TestHeaders:
    def test_multiple_values(self):
        """Test that headers with multiple values are parsed correctly."""
        support_mutliple_values_headers = [
            "Accept",
            "Accept-Encoding",
            "Accept-Language",
            "Call-Info",
            "Allow",
            "Contact",
            "Content-Encoding",
            "Content-Language",
            "Error-Info",
            "In-Reply-To",
            "Proxy-Require",
            "Record-Route",
            "Require",
            "Route",
            "Supported",
            "Unsupported",
            "Via",
            "Warning",
        ]
        wrong_classes = []
        for header_name in support_mutliple_values_headers:
            try:
                header_cls = Header.__registry_get_class_for__(header_name)
            except KeyError:  # noqa: PERF203
                _logger.info(f"Header {header_name} not implemented yet")
            else:
                if not issubclass(header_cls, MultipleValuesHeader):
                    wrong_classes.append(header_name)
        assert not wrong_classes, (
            "Headers with multiple values should be MultipleValuesHeader"
        )


class TestSIPMessages:
    def test_parse(self, sip_packets):
        """Test that all the sample SIP messages can be parsed without critical errors."""
        for packet in sip_packets:
            SIPMessage.parse(packet.data)
            # TODO: test origin
            # TODO: do asserts making sure (somehow) that parse->serialize == input

    def test_headers(self, sip_packets):
        """Test that all the sample SIP messages have the correct headers."""
        for packet in sip_packets:
            message = SIPMessage.parse(packet.data)
            headers = message.headers

            raw_headers_lines = (
                packet.data.decode().split("\r\n\r\n", 1)[0].split("\r\n")[1:]
            )
            raw_headers = dict(tuple(line.split(": ", 1)) for line in raw_headers_lines)
            assert list(headers.keys()) == list(raw_headers.keys()), (
                "Headers should be in the same order as the original message"
            )

            def clean(s):
                """Clean headers for comparison."""
                if not s.endswith("\r\n"):
                    s += "\r\n"
                # remove extra whitespace in each header, collapse to a single space
                s = re.sub(r"( |(?<!\r)\n)+", " ", s)
                # remove whitespace around commas
                s = re.sub(r"\s*,\s*", ",", s)
                # make all bool values uppercase
                s = re.sub(
                    r"\b(false|true)\b",
                    lambda m: m.group(1).upper(),
                    s,
                    flags=re.IGNORECASE,
                )
                # remove unnecessary quotes from nc=
                s = re.sub(r'\bnc="(\w+)"', r"nc=\1", s, flags=re.IGNORECASE)
                # add quotes when missing in display name before <sip:...>
                s = re.sub(r"(\w+) *(?=<sip:)", r'"\1" ', s)
                # strip leading and trailing whitespace in each line
                s = re.sub(r"^ +| +$", "", s, flags=re.MULTILINE)
                return s.strip()

            assert clean(str(headers)) == clean("\r\n".join(raw_headers_lines)), (
                "Headers should serialize to the same string as the original message"
            )

            previous_headers = Headers()
            for header in headers.values():
                hdr_cls_name = header.__class__.__name__
                assert header.name in headers, (
                    f"{hdr_cls_name}: header name should be in Headers map"
                )
                assert (  # noqa: PT018
                    header.name.upper() in headers and header.name.lower() in headers
                ), f"{hdr_cls_name}: headers should be case-insensitive"

                serialized_value = header.serialize()
                rebuilt_header = Header.parse(
                    header.name, serialized_value, previous_headers
                )
                assert rebuilt_header.serialize() == serialized_value, (
                    f"{hdr_cls_name}: value should serialize without loss"
                )
                assert str(rebuilt_header) == str(header), (
                    f"{hdr_cls_name}: entire header should serialize without loss"
                )
                assert rebuilt_header == header, (
                    f"{hdr_cls_name}: should be able to be rebuilt and still match"
                )

                previous_headers[header.name] = rebuilt_header


class AuthParams(NamedTuple):
    realm: str
    nonce: str
    method: str
    qop: str | None
    login: str
    password: str
    auth_uri: SIPURI
    cnonce: str | None = None
    nc: int | None = None


@pytest.fixture
def auth_samples() -> list[tuple[AuthParams, HashedAuth]]:
    auth_responses = [
        (
            AuthParams(
                realm="sip.twilio.com",
                nonce="3KmK8fSJKaapgVlXobV26UmcL8Qq1WqRjl_UQX_Pmey2nmI_",
                method="REGISTER",
                qop="auth",
                login="wa-bot-01",
                password="HunterHunter2",
                auth_uri=SIPURI.parse(
                    "sip:wa-test.sip.frankfurt.twilio.com;transport=UDP"
                ),
                cnonce="a2d67228",
                nc=14099,
            ),
            "3fa72cc5973eb4d986d7529503ae3b0c",
        )
    ]
    return [
        (
            auth,
            HashedAuth(
                username=auth.login,
                realm=auth.realm,
                nonce=auth.nonce,
                qop=auth.qop,
                nc=f"{auth.nc:08}" if isinstance(auth.nc, int) else str(auth.nc),
                cnonce=auth.cnonce,
                uri=str(auth.auth_uri),
                response=response,
                algorithm="MD5",
            ),
        )
        for auth, response in auth_responses
    ]


def test_auth_hash(auth_samples):
    for auth, expected in auth_samples:
        result = compute_auth(**auth._asdict())
        assert result == expected


PacketAndSIPMessage = namedtuple("PacketAndSIPMessage", ["packet", "message"])


@pytest.fixture
def sip_transactions(sip_packets):
    """
    Return lists of packets and SIP messages, grouped by transaction.
    Duplicate messages are removed (eg. retransmissions).
    """
    transactions = defaultdict(list)
    for packet in sip_packets:
        sip_message = SIPMessage.parse(packet.data)  # TODO: add origin
        call_id = sip_message.headers.get("Call-ID")
        if call_id is None:
            _logger.debug("SIP message has no Call-ID header, skipping")
        transactions[str(call_id)].append(PacketAndSIPMessage(packet, sip_message))

    # use a sliding window of 10 messages to remove duplicates
    for packets in transactions.values():
        seen = deque(maxlen=10)
        packets[:] = [
            pm
            for pm in packets
            if pm.message not in seen and (not seen.append(pm.message))
        ]

    return transactions


@pytest.fixture
def sip_registrations(sip_transactions):
    """Return lists of SIP REGISTER transactions, grouped by transaction."""
    return {
        call_id: messages
        for call_id, messages in sip_transactions.items()
        if messages[0].message.method == SIPMethod.REGISTER
    }


@pytest.fixture
def sip_invites(sip_transactions):
    """Return lists of SIP INVITE transactions, grouped by transaction."""
    return {
        call_id: messages
        for call_id, messages in sip_transactions.items()
        if messages[0].message.method == SIPMethod.INVITE
    }


@pytest.fixture
def incoming_invites(sip_invites):
    """Return lists of incoming SIP INVITE transactions, grouped by transaction."""
    return {
        call_id: messages
        for call_id, messages in sip_invites.items()
        if messages[0].packet.dest == Dest.CLIENT
    }


@pytest.fixture
def outgoing_invites(sip_invites):
    """Return lists of outgoing SIP INVITE transactions, grouped by transaction."""
    return {
        call_id: messages
        for call_id, messages in sip_invites.items()
        if messages[0].packet.dest == Dest.SERVER
    }


class MockSIPServer(MockServer[PacketAndSIPMessage]):
    """Small mock SIP server, opens up a connection and sends/recvs data."""

    def __init__(self, *args, wait_recv_timeout: float = 10.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_sent_msg: SIPMessage | None = None
        self.last_recv_msg: SIPMessage | None = None
        self.wait_recv_timeout: float = wait_recv_timeout
        self.recv_queue = deque()
        self.sent_count = 0
        self.recv_count = 0

    def send(self, packet: PacketAndSIPMessage):
        assert self.socket is not None

        if packet is None:
            _logger.debug(f"Waiting for recv, ready responses: {len(self.recv_queue)}")
        start_time = time.time()
        while packet is None and not self.recv_queue:
            time.sleep(1e-3)
            if self.stop_event.is_set():
                return
            if time.time() - start_time > self.wait_recv_timeout:
                raise TimeoutError("Timed out waiting for recv")
        if packet is None:
            self.recv_queue.popleft()
            return

        message = packet.message

        if self.last_recv_msg is not None and isinstance(
            self.last_recv_msg, SIPRequest
        ):
            # adapt transaction ID, cseq, from and to
            message.headers["Call-ID"] = self.last_recv_msg.headers["Call-ID"]
            message.headers["CSeq"] = self.last_recv_msg.headers["CSeq"]
            message.headers["From"] = self.last_recv_msg.headers["From"]
            message.headers["To"] = self.last_recv_msg.headers["To"]

        self.recv_queue.clear()
        self.socket.sendto(bytes(message), self.client_address)
        self.last_sent_msg = message
        self.sent_count += 1
        _logger.debug(f"Sent {message!r}")

    def recv(self):
        try:
            data, addr = self.socket.recvfrom(8192)
        except (TimeoutError, BlockingIOError):
            pass
        else:
            if not data.strip():
                return
            message = SIPMessage.parse(data, origin=addr)
            self.last_recv_msg = message
            self.recv_count += 1
            self.recv_queue.append(message)
            _logger.debug(f"Received {message!r}")


class TestCallHandler(CallHandler):
    """A mock VoIP phone that can send and receive SIP messages."""

    @property
    def can_accept_calls(self) -> bool:
        return True

    @property
    def can_make_calls(self) -> bool:
        return True

    def prepare_call(self, call: SIPCall) -> None:
        pass

    def teardown_call(self, call: SIPCall) -> None:
        pass

    async def answer(self, call: SIPCall) -> bool:
        return True

    def get_rtp_profiles_by_port(self) -> Mapping[int, Sequence[rtp.RTPMediaProfiles]]:
        return {
            5060: [rtp.RTPMediaProfiles.PCMU, rtp.RTPMediaProfiles.PCMA],
        }

    def get_media_flow(self) -> rtp.MediaFlowType:
        return rtp.MediaFlowType.SENDRECV

    def establish_call(self, call: SIPCall) -> None:
        pass

    def terminate_call(self, call: SIPCall) -> None:
        pass

    def on_call_failure(self, call: SIPCall, error: Exception) -> bool | None:
        pass


def mute_caplog(caplog, mute, logger_name=None):
    """Return a context manager to maybe mute the caplog for the given logger."""
    if mute:
        log_level_context = caplog.at_level(logging.CRITICAL, logger=logger_name)
    else:
        log_level_context = nullcontext()
    return log_level_context


@pytest.fixture
def _skip_register(monkeypatch):
    async def mock_register(self):  # noqa: RUF029
        self._registered = True

    monkeypatch.setattr(SIPRegistration, "register", mock_register)


@pytest.fixture
def _skip_deregister(monkeypatch):
    original__register_transaction = SIPRegistration._register_transaction

    async def mock_register_transaction(self, *, deregister: bool = False) -> None:
        if deregister:
            return None  # skip deregister
        return await original__register_transaction(self, deregister=deregister)

    monkeypatch.setattr(
        SIPRegistration, "_register_transaction", mock_register_transaction
    )


class TestSIPClient:
    """Test SIPClient class."""

    @classmethod
    @contextmanager
    def _init_server_client(
        cls,
        server_packets,
        *,
        server_address=None,
        client_address=None,
        wait_recv_timeout=2.0,
        register_timeout=2e-1,
        register_expires=1,
        default_response_timeout=2e-1,
    ):
        if server_address is None:
            server_address = "127.0.0.1", 0
        if client_address is None:
            client_address = "127.0.0.1", 0

        server = MockSIPServer(
            iter(server_packets),
            server_address,
            client_address,
            send_delay=1e-2,
            wait_recv_timeout=wait_recv_timeout,
            pre_bind=True,
        )
        server_address = server.socket.getsockname()
        client = SIPClient(
            call_handler_factory=lambda call: TestCallHandler(),  # noqa: ARG005
            username="alice",
            password="secret",
            server_host=server_address[0],
            server_port=server_address[1],
            local_host=client_address[0],
            local_port=client_address[1],
            register_timeout=register_timeout,
            register_expires=register_expires,
            default_response_timeout=default_response_timeout,
            keep_alive_interval=None,  # disable keep-alive
            pre_bind=True,
        )
        server.client_address = client.local_addr

        yield server, client

    @classmethod
    def _test_registration(cls, server_packets, expect_failure, **kwargs):
        with cls._init_server_client(server_packets, **kwargs) as (server, client):  # noqa: SIM117
            with server:
                with pytest.raises(Exception) as exc_info, client:
                    while not server.send_done and not client.closed:
                        time.sleep(1e-9)
                    _logger.debug("Stopping test")

                    assert expect_failure or client.registered, (
                        "Client should be registered"
                    )

                    raise StopIteration

                assert exc_info.type is StopIteration or (
                    issubclass(exc_info.type, SIPException) and expect_failure
                ), (
                    f"SIP client failed with unexpected {exc_info.type}: {exc_info.value}"
                    f"\n{traceback.format_tb(exc_info.tb)}"
                    f"\nlast sent message: {server.last_sent_msg!r}"
                    f"\nlast received message: {server.last_recv_msg!r}"
                )

    @pytest.mark.usefixtures("_skip_deregister")
    def test_register(self, sip_registrations, monkeypatch, caplog):
        """Test that the client can register with the server."""
        for call_id, packets in sip_registrations.items():
            server_packets = [
                pm if pm.packet.dest == Dest.CLIENT else None for pm in packets
            ]
            # check if we have any 400,402+ responses in the server packets or 2 or more
            # subsequent 401, and set an expect_failure flag.
            # We have only REGISTER packets here, so we can skip the CSeq check.
            expect_failure = False
            last_message = None
            for pm in server_packets:
                if pm is None:
                    continue
                if (
                    (
                        isinstance(pm.message, SIPResponse)
                        and pm.message.status.code == 400
                    )
                    or pm.message.status.code >= 402
                    or (
                        pm.message.status.code == 401
                        and last_message
                        and isinstance(last_message, SIPResponse)
                        and last_message.status.code == 401
                    )
                ):
                    expect_failure = True
                    break
                last_message = pm.message

            try:
                with mute_caplog(caplog, expect_failure, "sibilant.sip.client"):
                    self._test_registration(server_packets, expect_failure)
            except Exception:
                print(f"{self.test_register.__name__} failed for call ID: {call_id}")
                raise

    @classmethod
    def _test_invite_wrapper(
        cls, call_test_fn, server_packets, expected_states, **kwargs
    ):
        with cls._init_server_client(server_packets, **kwargs) as (server, client):  # noqa: SIM117
            with client, server:
                call = call_test_fn(client, server)

                assert not client._pending_futures, "expected client to be done"
                assert server.sent_count, "at least one message should have been sent"
                assert call, "expected at least one call to have started"
                assert call.state in expected_states, (
                    "call should be in expected states"
                )
                expected_sent_count = len([m for m in server_packets if m is not None])
                assert server.sent_count == expected_sent_count

    @classmethod
    def _test_invite_incoming(cls, server_packets, expected_states, **kwargs):
        def grab_call_and_wait(client, server):
            call = None
            while (
                not server.send_done
                or client._pending_futures  # FIXME: might get stuck on keepalive
            ) and not server.error:
                if call is None and client.calls:
                    call = next(iter(client.calls.values()))
                time.sleep(1e-9)
            time.sleep(1e-1)
            if server.error:
                client.stop()
                raise server.error
            return call

        return cls._test_invite_wrapper(
            grab_call_and_wait, server_packets, expected_states, **kwargs
        )

    @pytest.mark.usefixtures("_skip_register", "_skip_deregister")
    def test_invite_incoming(self, incoming_invites, caplog):
        """Test that the client can send an INVITE to the server."""
        for call_id, packets in incoming_invites.items():
            # the flow is >INVITE, <180 Ringing, <200 Ok, >ACK [, >BYE, <200 Ok]
            # so make sure there's the correct order in server packets
            server_packets = []
            last_method = None
            for pm in packets:
                if pm.packet.dest == Dest.SERVER:
                    continue
                msg = pm.message
                if not server_packets and msg.method == SIPMethod.INVITE:
                    server_packets.extend([pm, None, None])  # add wait for 180 and 200
                elif last_method == SIPMethod.INVITE and msg.method == SIPMethod.ACK:
                    server_packets.append(pm)
                elif last_method == SIPMethod.ACK and msg.method == SIPMethod.BYE:
                    server_packets.extend([pm, None])  # add wait for final 200
                else:
                    _logger.warning(f"Unexpected packet in incoming INVITE flow: {pm}")
                    server_packets = []
                    break
                last_method = msg.method if msg else None
            if not server_packets:
                continue

            if last_method == SIPMethod.INVITE:
                expected_states = (CallState.FAILED,)  # never replied
            elif last_method == SIPMethod.ACK:
                expected_states = (CallState.ESTABLISHED,)
            elif last_method == SIPMethod.BYE:
                expected_states = (CallState.HUNG_UP,)
            else:
                raise RuntimeError(
                    f"Unexpected last method in incoming INVITE flow: {last_method}"
                )

            expect_failure = CallState.FAILED in expected_states
            try:
                with mute_caplog(caplog, expect_failure, "sibilant.sip.client"):
                    self._test_invite_incoming(server_packets, expected_states)
            except Exception:
                print(
                    f"{self.test_invite_incoming.__name__} failed for call ID: {call_id}"
                )
                raise

    @classmethod
    def _test_invite_outgoing(
        cls, contact, client_methods, server_packets, expected_states, **kwargs
    ):
        def send_invite_and_wait(client, server):
            # TODO: more client methods to simulate call flow (ignore ACK)
            call = client.invite(contact)
            while not server.send_done or client._pending_futures:
                time.sleep(1e-9)
            time.sleep(1e-1)
            return call

        return cls._test_invite_wrapper(
            send_invite_and_wait, server_packets, expected_states, **kwargs
        )

    @pytest.mark.usefixtures("_skip_register", "_skip_deregister")
    def test_invite_outgoing(self, outgoing_invites, caplog):
        """Test that the client can send an INVITE to the server."""
        for call_id, packets in outgoing_invites.items():
            assert isinstance(packets[0].message, SIPRequest)
            assert packets[0].message.method == SIPMethod.INVITE
            contact = packets[0].message.uri

            server_packets = []
            last_response_status = None
            client_methods = []
            for pm in packets:
                # calls communication can happen between client and client
                if isinstance(pm.message, SIPRequest):
                    method = pm.message.method
                    if method == SIPMethod.INVITE and method in client_methods:
                        break  # TODO: better split retransmitted packets from server
                    if method in {SIPMethod.CANCEL, SIPMethod.BYE}:
                        break  # TODO: implement
                    client_methods.append(method)
                    server_packets.append(None)
                elif isinstance(pm.message, SIPResponse):
                    server_packets.append(pm)
                    last_response_status = pm.message.status
                    if last_response_status.code not in {100, 180, 200}:
                        break
                else:
                    raise TypeError(f"Unexpected packet: {pm}")

            if last_response_status.code == 200:
                expected_states = (CallState.ESTABLISHED,)
            else:
                expected_states = (CallState.FAILED,)

            expect_failure = CallState.FAILED in expected_states

            try:
                with mute_caplog(caplog, expect_failure, "sibilant.sip.client"):
                    self._test_invite_outgoing(
                        contact, client_methods, server_packets, expected_states
                    )
            except Exception:
                print(
                    f"{self.test_invite_outgoing.__name__} failed for call ID: {call_id}"
                )
                raise



"""Common SIP / SDP / RTP structures."""

from __future__ import annotations

import re
from dataclasses import field as dataclass_field
from re import Match
from typing import TYPE_CHECKING

from frozendict import frozendict
from typing_extensions import Self

from .exceptions import SIPParseError
from .helpers import ParseableSerializable, slots_dataclass


if TYPE_CHECKING:
    from collections.abc import Collection, Mapping


DEFAULT_SCHEME: str = "sip"

UNRESERVED_C: str = r"_.!~*'()%\-"
USER_C: str = rf"[\w{UNRESERVED_C}+\-&$,;?/]"
PWD_C: str = rf"[\w{UNRESERVED_C}+\-&$,]"
PARAM_C: str = rf"[\w{UNRESERVED_C}\[\]/:&+$]"
HEADER_C: str = rf"[\w{UNRESERVED_C}\[\]/:?+$]"

DISPLAY_NAME_PAT: str = r"(?P<display_name>[\"'][^\"']+[\"']|\S.*?)(?=\s*\<)"
CONTACT_PAT: str = rf"(?P<user>{USER_C}+)(?::(?P<password>{PWD_C}+))?(?=@)"
IPv4_D_PAT: str = r"(?:1?\d{1,2}|2[0-4]\d|25[0-5])"
IPv4_PAT: str = rf"(?:(?:{IPv4_D_PAT}\.){{3}}{IPv4_D_PAT})"
# FIXME: IPv6
DNS_LABEL_PAT: str = r"(?i:[a-z0-9]([a-z-0-9-]{0,61}[a-z0-9])?)"
FQDN_PAT: str = rf"(?:(?:{DNS_LABEL_PAT}\.)*{DNS_LABEL_PAT}\.?)"
HOSTNAME_PAT: str = rf"(?P<hostname>{FQDN_PAT}|{IPv4_PAT})"
HOST_PAT: str = rf"(?P<host>{HOSTNAME_PAT})(?::(?P<port>\d+))?"
SCHEME_PAT: str = r"(?P<scheme>sips?)(?=:)"
PARAMS_PAT: str = rf"(?P<params>(?:;{PARAM_C}+(?:={PARAM_C}+)?)+)"
HEADERS_PAT: str = rf"(?=\?)(?P<headers>(?:[?&]{USER_C}+={USER_C}+)+)"
URI_PART_PAT: str = (
    rf"(?:{SCHEME_PAT}:)?(?:{CONTACT_PAT}@)?{HOST_PAT}"
    rf"(?:{PARAMS_PAT})?(?:{HEADERS_PAT})?"
)
URI_PATS: Collection[str] = [URI_PART_PAT, rf"<{URI_PART_PAT}>"]
ADDRESS_PATS: Collection[str] = [
    rf"(?:{DISPLAY_NAME_PAT}\s*)?(?P<uri>{p})" for p in URI_PATS
]


@slots_dataclass(frozen=True)
class SIPURI(ParseableSerializable):
    """A SIP URI."""

    host: str
    port: int | None = None
    user: str | None = None
    password: str | None = None
    scheme: str = DEFAULT_SCHEME
    params: Mapping[str, str | None] = dataclass_field(default_factory=frozendict)
    headers: Mapping[str, str] = dataclass_field(default_factory=frozendict)

    brackets: bool = False

    @classmethod
    def parse(cls, value: str, *, force_brackets: bool | None = None) -> SIPURI:
        """Parse a SIP URI."""
        match: Match | None = None
        for uri_pat in URI_PATS:
            if match := re.fullmatch(uri_pat, value.strip()):
                break
        if match is None:
            raise SIPParseError(f"Invalid SIP URI: {value}")
        if force_brackets is None:
            brackets = value.strip().startswith("<") and value.strip().endswith(">")
        else:
            brackets = force_brackets
        params: dict[str, str | None] = {}
        if params_raw := match.group("params"):
            params = {
                name: value
                for param in params_raw.split(";")
                if param
                for name, value in (
                    param.split("=") if "=" in param else (param, None),
                )
            }
        headers: dict[str, str] = {}
        if headers_raw := match.group("headers"):
            headers = {
                name: value
                for header in headers_raw.lstrip("?").split("&")
                if header
                for name, value in header.split("=")
            }
        return cls(
            host=match.group("host"),
            port=match.group("port"),
            user=match.group("user"),
            password=match.group("password"),
            scheme=match.group("scheme") or DEFAULT_SCHEME,
            params=frozendict(params),
            headers=frozendict(headers),
            brackets=brackets,
        )

    def serialize(self, *, force_brackets: bool | None = None) -> str:
        """Serialize the SIP URI to a string."""
        password: str = f":{self.password}" if self.password else ""
        login: str = f"{self.user}{password}@" if self.user else ""
        hostname: str = f"{self.host}:{self.port}" if self.port else self.host
        params: str = "".join(
            f";{name}={value}" if value is not None else f";{name}"
            for name, value in self.params.items()
        )
        headers: str = "".join(
            f"?{name}={value}" for name, value in self.headers.items()
        )
        brackets: bool = self.brackets if force_brackets is None else force_brackets
        uri: str = f"{self.scheme}:" + login + hostname + params + headers
        if brackets:
            uri = f"<{uri}>"
        return uri

    def __str__(self) -> str:
        return self.serialize()


@slots_dataclass(frozen=True)
class SIPAddress:
    """A SIP contact address, with an optional display name and a SIP URI."""

    uri: SIPURI
    display_name: str | None = None

    @classmethod
    def parse(cls, value: str, *, force_brackets: bool | None = None) -> Self:
        """Parse a SIP address from a string. Optionally with a display name and phone number."""
        match: Match | None = None
        for address_pat in ADDRESS_PATS:
            if match := re.fullmatch(address_pat, value):
                break
        match_groups: dict[str, str] = match.groupdict() if match else {}
        display_name = match_groups.get("display_name")
        if display_name and display_name[0] in {"'", '"'}:
            display_name = re.sub(r"^\s*([\"'])(.*?)\1\s*$", r"\2", display_name)
        uri_raw: str | None = match_groups.get("uri") if match else None
        if not match or not uri_raw:
            raise SIPParseError(f"Invalid SIP address: {value}")
        assert uri_raw is not None
        if force_brackets is None:
            force_brackets = bool(match_groups)
        uri: SIPURI = SIPURI.parse(uri_raw, force_brackets=force_brackets)
        return cls(uri=uri, display_name=display_name)

    def __str__(self) -> str:
        """Serialize the SIP address to a string."""
        if self.display_name:
            return f'"{self.display_name}" {self.uri.serialize(force_brackets=True)}'
        else:
            return str(self.uri)



"""Classes for RTP packets and related utilities."""

from __future__ import annotations

import enum
import time
from contextlib import contextmanager
from dataclasses import (
    field as dataclass_field,
    fields as dataclass_fields,
    replace as dataclass_replace,
)
from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, Any, ClassVar, cast

from cbitstruct import CompiledFormat
from typing_extensions import Self

from sibilant.codecs import Codec, PCMACodec, PCMUCodec
from sibilant.constants import SUPPORTED_RTP_VERSIONS
from sibilant.exceptions import RTPUnsupportedCodec, RTPUnsupportedVersion
from sibilant.helpers import (
    DataclassEnum,
    FieldsEnum,
    FieldsEnumDatatype,
    ParseableSerializableRaw,
    slots_dataclass,
)


if TYPE_CHECKING:
    from collections.abc import Iterator

    import numpy as np
    from numpy.typing import NDArray


class RTPMediaType(enum.Enum):
    """The media types of RTP media format."""

    AUDIO = "audio"
    VIDEO = "video"
    AUDIO_VIDEO = "audio,video"


@slots_dataclass
class RTPMediaFormat(FieldsEnumDatatype):
    """Represents an RTP media format type."""

    # FIXME: should this be int only?
    payload_type: int | str
    media_type: RTPMediaType
    encoding_name: str
    clock_rate: int
    channels: int | None = None
    format_specific_parameters: str | None = None

    codec_type: type[Codec] | None = None
    codec: Codec | None = None

    def __post_init__(self) -> None:
        if self.codec is None and self.codec_type is not None:
            self.codec = self.codec_type()
        elif self.codec_type is None:
            self.codec_type = type(self.codec) if self.codec is not None else None

    @property
    def enum_value(self) -> Any:  # noqa: D102
        return self.payload_type

    @property
    def mimetype(self) -> str:  # TODO: is this correct?
        """The mimetype of the media format."""
        return f"{self.media_type.value}/{self.encoding_name}".lower()

    def encode(self, data: NDArray[np.float32]) -> bytes:
        """Encode float32 audio in the [-1, 1] range to bytes."""
        if self.codec is None:
            raise RTPUnsupportedCodec(f"No codec set for this media format: {self!r}")
        return self.codec.encode(data)

    def decode(self, data: bytes) -> NDArray[np.float32]:
        """Decode bytes to float32 audio in the [-1, 1] range."""
        if self.codec is None:
            raise RTPUnsupportedCodec(f"No codec set for this media format: {self!r}")
        return self.codec.decode(data)

    def __hash__(self) -> int:
        return hash((
            self.payload_type,
            self.media_type,
            self.encoding_name,
            self.clock_rate,
            self.channels,
            self.format_specific_parameters,
        ))


UNKNOWN_FORMAT = RTPMediaFormat("unknown", RTPMediaType.AUDIO_VIDEO, "unknown", 0)

# TODO: implement enum matching on multiple fields (i.e. composite _value_), using a separate cls var


class RTPMediaProfiles(RTPMediaFormat, DataclassEnum):
    """The known RTP media profiles."""

    __wrapped_type__ = RTPMediaFormat
    __allow_unknown__ = True

    @classmethod
    @lru_cache(maxsize=256)
    def match(
        cls,
        payload_type: int | str | None = None,
        media_format: RTPMediaFormat | None = None,
    ) -> RTPMediaProfiles:
        """
        Tries to match a media format to a profile.
        Will try to match the payload type first, then the media format,
        then try to match the encoding name as payload type.
        In case no match is found, the unknown profile is returned.

        :param payload_type: The payload type to match.
        :param media_format: The media format to match.
        :return: The matched or unknown profile.
        """
        payload_type_raw: int | str
        if payload_type is not None:
            payload_type_raw = payload_type
        elif media_format is not None:
            payload_type_raw = media_format.payload_type
        else:
            raise ValueError("One of `payload_type` or `media_format` must be provided")

        media_profile: RTPMediaProfiles | None = None
        try:
            media_profile = RTPMediaProfiles.match_value(payload_type_raw)

        except (TypeError, ValueError):
            if isinstance(payload_type_raw, str):
                for known_media_profile in RTPMediaProfiles:
                    if known_media_profile.encoding_name == payload_type_raw:
                        media_profile = known_media_profile
                        break

        if media_profile is None:
            media_format = media_format or dataclass_replace(
                UNKNOWN_FORMAT, payload_type=payload_type_raw
            )
            media_profile = object.__new__(cls)
            media_profile._name_ = "UNKNOWN"
            media_profile._value_ = media_format
            media_profile.__objclass__ = cls  # type: ignore[attr-defined]
            media_profile._sort_order_ = 9999  # type: ignore[attr-defined]
            media_profile.__init__(  # type: ignore[misc]  # noqa: PLC2801
                *(getattr(media_format, f.name) for f in dataclass_fields(media_format))
            )
        assert media_profile is not None
        return media_profile

    @cached_property
    def fmt(self) -> RTPMediaFormat:
        """The wrapped media format."""
        if isinstance(self, FieldsEnum):
            return cast("RTPMediaFormat", self._wrapped_value_)
        else:
            return cast("RTPMediaFormat", self._value_)

    # audio
    PCMU = (0, RTPMediaType.AUDIO, "PCMU", 8000, 1, None, PCMUCodec)
    GSM = (3, RTPMediaType.AUDIO, "GSM", 8000, 1)
    G723 = (4, RTPMediaType.AUDIO, "G723", 8000, 1)
    DVI4_8000 = (5, RTPMediaType.AUDIO, "DVI4", 8000, 1)
    DVI4_16000 = (6, RTPMediaType.AUDIO, "DVI4", 16000, 1)
    LPC = (7, RTPMediaType.AUDIO, "LPC", 8000, 1)
    PCMA = (8, RTPMediaType.AUDIO, "PCMA", 8000, 1, None, PCMACodec)
    G722 = (9, RTPMediaType.AUDIO, "G722", 8000, 1)
    L16_2 = (10, RTPMediaType.AUDIO, "L16", 44100, 2)
    L16 = (11, RTPMediaType.AUDIO, "L16", 44100, 1)
    QCELP = (12, RTPMediaType.AUDIO, "QCELP", 8000, 1)
    CN = (13, RTPMediaType.AUDIO, "CN", 8000, 1)
    MPA = (14, RTPMediaType.AUDIO, "MPA", 90000, None)
    G728 = (15, RTPMediaType.AUDIO, "G728", 8000, 1)
    DVI4_11025 = (16, RTPMediaType.AUDIO, "DVI4", 11025, 1)
    DVI4_22050 = (17, RTPMediaType.AUDIO, "DVI4", 22050, 1)
    G729 = (18, RTPMediaType.AUDIO, "G729", 8000, 1)
    # video
    CELLB = (25, RTPMediaType.VIDEO, "CELLB", 90000, None)
    JPEG = (26, RTPMediaType.VIDEO, "JPEG", 90000, None)
    NV = (28, RTPMediaType.VIDEO, "nv", 90000, None)
    H261 = (31, RTPMediaType.VIDEO, "H261", 90000, None)
    MPV = (32, RTPMediaType.VIDEO, "MPV", 90000, None)
    MP2T = (33, RTPMediaType.AUDIO_VIDEO, "MP2T", 90000, None)
    H263 = (34, RTPMediaType.VIDEO, "H263", 90000, None)
    # misc
    TELEPHONE_EVENT = (
        "telephone-event",
        RTPMediaType.AUDIO,
        "telephone-event",
        8000,
        None,
    )


@slots_dataclass
class RTPPacket(ParseableSerializableRaw):
    """Implements RTP packets as defined in :rfc:`3550#section-5.1`."""

    # header
    version: int
    padding: bool
    extension: bool
    csrc_count: int
    marker: bool
    payload_type: RTPMediaProfiles
    sequence: int
    timestamp: int
    ssrc: int
    csrc: list[int] = dataclass_field(default_factory=list)
    ext_id: int = 0
    ext_len: int = 0
    ext_data: bytes = b""
    # data
    payload: bytes = b""

    _duration_override: float | None = None

    _format_u64: ClassVar[CompiledFormat] = CompiledFormat("u2b1b1u4b1u7u16u32u32")
    _ext_header_u32: ClassVar[CompiledFormat] = CompiledFormat("u16u16")

    @classmethod
    def calc_header_len(cls, csrc_count: int, extension: bool, ext_len: int) -> int:  # noqa: FBT001
        """Calculate the length of the header in bytes."""
        extension_len: int = (
            cls._ext_header_u32.calcsize() // 8 + ext_len * 4 if extension else 0
        )
        return (
            cast("int", cls._format_u64.calcsize()) // 8
            + csrc_count * 4
            + extension_len
        )

    @property
    def duration(self) -> float:
        """Duration of the packet in seconds. Supports only PCMU and PCMA."""
        if self._duration_override is not None:
            return self._duration_override
        if self.payload_type in {RTPMediaProfiles.PCMU, RTPMediaProfiles.PCMA}:
            return len(self.payload) / self.payload_type.clock_rate
        raise NotImplementedError

    @classmethod
    def parse(cls, data: bytes) -> Self:  # noqa: PLR0914, D102
        (
            version,
            padding,
            extension,
            csrc_count,
            marker,
            payload_type_raw,
            sequence,
            timestamp,
            ssrc,
        ) = cls._format_u64.unpack(data)

        if version not in SUPPORTED_RTP_VERSIONS:
            raise RTPUnsupportedVersion(f"Unsupported RTP version: {version}")

        payload_type: RTPMediaProfiles = RTPMediaProfiles.match(payload_type_raw)

        csrc = []
        csrc_offset = cls._format_u64.calcsize() // 8
        if csrc_count:
            csrc_data = data[csrc_offset : csrc_offset + csrc_count * 4]
            csrc = [
                int.from_bytes(csrc_data[i : i + 4], "big")
                for i in range(0, len(csrc_data), 4)
            ]
        ext_id, ext_len, ext_data = 0, 0, b""
        if extension:
            ext_offset = csrc_offset + csrc_count * 4
            ext_header_len = cls._ext_header_u32.calcsize() // 8
            ext_header_raw = data[ext_offset : ext_offset + ext_header_len]
            ext_id, ext_len = cls._ext_header_u32.unpack(ext_header_raw)
            ext_data_offset = ext_offset + ext_header_len
            ext_data = data[ext_data_offset : ext_data_offset + ext_len * 4]

        payload_offset = cls.calc_header_len(csrc_count, extension, ext_len)
        payload = data[payload_offset:]

        return cls(
            version,
            padding,
            extension,
            csrc_count,
            marker,
            payload_type,
            sequence,
            timestamp,
            ssrc,
            csrc,
            ext_id,
            ext_len,
            ext_data,
            payload,
        )

    def serialize(self) -> bytes:  # noqa: D102
        header_data: bytes = cast(
            "bytes",
            self._format_u64.pack(
                self.version,
                self.padding,
                self.extension,
                self.csrc_count,
                self.marker,
                self.payload_type.payload_type,
                self.sequence,
                self.timestamp,
                self.ssrc,
            ),
        )
        if self.csrc_count:
            header_data += b"".join(int.to_bytes(csrc, 4, "big") for csrc in self.csrc)
        if self.extension:
            header_data += self._ext_header_u32.pack(self.ext_id, self.ext_len)
            header_data += self.ext_data
        return header_data + self.payload

    @property
    def header_len(self) -> int:
        """Length of the header in bytes."""
        return self.calc_header_len(self.csrc_count, self.extension, self.ext_len)

    def __len__(self) -> int:
        return self.header_len + len(self.payload)


class RTPPacketsStats:
    """An analytics utility class that collects statistics about RTP packets."""

    def __init__(self) -> None:
        self.count: int = 0
        self.bytes: int = 0
        self.duration: float = 0
        self.time: float = 0.0
        self.unimplemented: int = 0

    def add(self, packet: RTPPacket, elapsed: float) -> None:
        """Add a packet to the stats."""
        try:
            duration = packet.duration
        except NotImplementedError:
            self.unimplemented += 1
            return
        self.count += 1
        self.time += elapsed
        self.bytes += len(packet.payload)
        self.duration += duration

    @contextmanager
    def track(self, packet: RTPPacket) -> Iterator[Self]:
        """A context manager to track the time taken to process a packet."""
        start_time: int = time.perf_counter_ns()
        yield self
        end_time: int = time.perf_counter_ns()
        self.add(packet, (end_time - start_time) / 1e9)

    @property
    def count_per_sec(self) -> float:
        """Number of packets processed per second."""
        return self.count / self.time if self.time else 0

    @property
    def bytes_per_sec(self) -> float:
        """Number of bytes processed per second."""
        return self.bytes / self.time if self.time else 0

    @property
    def realtime_factor(self) -> float:
        """Realtime factor, i.e. how much faster than realtime the processing was."""
        return self.duration / self.time if self.time else 0

    def format(self) -> str:
        """Format the stats as a string."""
        return (
            f"{self.count} packets, {self.bytes_per_sec / (1024**2):,.2f} MB/s, "
            f"{self.count_per_sec:,.2f} packets/s, {self.realtime_factor:,.2f}x realtime"
        )

    def __bool__(self) -> bool:
        return self.time > 0

    def __add__(self, other: object) -> RTPPacketsStats:
        if not isinstance(other, RTPPacketsStats):
            return NotImplemented

        new = RTPPacketsStats()
        new.count = self.count + other.count
        new.bytes = self.bytes + other.bytes
        new.duration = self.duration + other.duration
        new.time = self.time + other.time
        new.unimplemented = self.unimplemented + other.unimplemented
        return new

    def __iadd__(self, other: object) -> Self:
        if not isinstance(other, RTPPacketsStats):
            return NotImplemented

        self.count += other.count
        self.bytes += other.bytes
        self.duration += other.duration
        self.time += other.time
        self.unimplemented += other.unimplemented
        return self



"""Base classes for audio codecs."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray


class Codec(ABC):
    """Abstract base class for audio codecs."""

    @abstractmethod
    def encode(self, data: NDArray[np.float32]) -> bytes:
        """Encode the given audio data."""

    @abstractmethod
    def decode(self, data: bytes) -> NDArray[np.float32]:
        """Decode the given audio data."""



from __future__ import annotations

import importlib.metadata as importlib_metadata
import warnings
from email.message import Message
from pathlib import Path
from typing import TYPE_CHECKING, Any

import toml


if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence


metadata: Message | importlib_metadata.PackageMetadata | Mapping[str, Any] | None = None
# FIXME: give precedence to pyproject.toml metadata for correct info in local testing
try:
    metadata = importlib_metadata.metadata(__package__ or __name__)
except importlib_metadata.PackageNotFoundError:
    init_path = Path(__file__).resolve().parent
    for relpaths in (("..", "pyproject.toml"), ("pyproject.toml",)):
        pyproj_toml_path = Path(init_path, *relpaths)
        if pyproj_toml_path.exists():
            metadata = toml.load(pyproj_toml_path)
            break
    else:
        warnings.warn(
            "Didn't find distinfo nor pyproject.toml for package metadata", stacklevel=1
        )


def get_metadata(
    distinfo_key: str,
    toml_getter: str | int | Sequence[str | int] | Callable[[Mapping[str, Any]], Any],
) -> Any:
    global metadata
    if metadata is None:
        return None
    if isinstance(metadata, Message):
        return metadata.get(distinfo_key)
    elif isinstance(metadata, dict):
        try:
            if callable(toml_getter):
                return toml_getter(metadata)
            elif isinstance(toml_getter, (list, tuple)):
                value = metadata
                for key in toml_getter:
                    value = value[key]
                return value
            else:
                return metadata[toml_getter]
        except (KeyError, IndexError):
            return None
    return None



"""Various constants used by the sibilant library."""

from __future__ import annotations

import re as _re
import typing as _typing


SUPPORTED_SIP_VERSIONS: list[str] = ["SIP/2.0"]
SUPPORTED_SDP_VERSIONS: list[str] = ["0"]
SUPPORTED_RTP_VERSIONS: list[int] = [2]
SUPPORTED_RTP_PROFILES: list[str] = ["PCMU", "PCMA", "telephone-event"]

DEFAULT_SIP_PORT: int = 5060
DEFAULT_RTP_PORT_RANGE: tuple[int, int] = (6000, 7000)

PUBLIC_IP_RESOLVERS: list[tuple[str, _typing.Callable[[str], str | None] | None]] = [
    (
        "http://cloudflare.com/cdn-cgi/trace",
        lambda body: (m := _re.search(r"ip=(.+?)(?=\s|[\r\n]|$)", body)) and m.group(1),
    ),
    ("http://ident.me", None),
    ("http://ifconfig.me/ip", None),
    ("http://icanhazip.com/", None),
    (
        "http://checkip.dyndns.org/",
        lambda body: (
            m := _re.search(r"ip address: ?(.*?)\b", body, flags=_re.IGNORECASE)
        )
        and m.group(1),
    ),
    (
        "https://cloudflare.com/cdn-cgi/trace",
        lambda body: (m := _re.search(r"ip=(.+?)(?=\s|[\r\n]|$)", body)) and m.group(1),
    ),
    ("https://ident.me", None),
    ("https://ifconfig.me/ip", None),
    ("https://icanhazip.com/", None),
]



"""Exception calsses for the Sibilant library."""

from __future__ import annotations

from typing import Any


class SibilantException(Exception):
    """Base class for all custom library exceptions."""


class ParseError(SibilantException, ValueError):
    """Raised when a packet cannot be parsed."""


class SIPException(SibilantException):
    """Base class for all exceptions raised by the SIP module."""


class SIPMessageException(SIPException, IOError):
    """Exceptions related to SIP messages."""

    def __init__(self, *args: Any, **kwargs: Any):
        """Initialize SIPRequestException with `request` and `response` objects."""
        response = kwargs.pop("response", None)
        self.response = response
        self.request = kwargs.pop("request", None)
        if response is not None and not self.request and hasattr(response, "request"):
            self.request = self.response.request
        super().__init__(*args, **kwargs)


class SIPUnsupportedError(SIPMessageException, NotImplementedError):
    """Exception raised when a SIP message or feature is not supported."""


class SIPUnsupportedVersion(SIPUnsupportedError):
    """The SIP version is not supported by this library."""


class SIPParseError(SIPException, ParseError):
    """Exceptions related to SIP messages / data parsing."""


class SIPTimeout(SIPException, TimeoutError):
    """Raised when a SIP transaction times out."""


class SIPBadMessage(SIPException):
    """Raised when a SIP message is malformed or invalid."""


class SIPBadRequest(SIPBadMessage):
    """Raised when a SIP request is malformed or invalid."""


class SIPBadResponse(SIPBadMessage):
    """Raised when a SIP response is malformed or invalid."""


class SIPAuthenticationError(SIPException):
    """Raised when a SIP authentication fails."""


class SDPException(SibilantException):
    """Base class for all exceptions raised by the SDP module."""


class SDPUnsupportedVersion(SDPException, NotImplementedError):
    """The SDP version is not supported by this library."""


class SDPParseError(SDPException, ParseError):
    """Exception related to SDP data parsing."""


class SDPUnknownFieldError(SDPParseError):
    """Exception raised when an unknown SDP field is encountered."""


class RTPException(SibilantException):
    """Base class for all exceptions raised by the RTP module."""


class RTPUnsupportedVersion(SIPUnsupportedError):
    """The RTP version is not supported by this library."""


class RTPParseError(RTPException, ParseError):
    """Exception related to RTP data parsing."""


class RTPMismatchedStreamError(RTPException):
    """Exception raised when a packet does not belong to the stream."""


class RTPBrokenStreamError(RTPException):
    """Exception raised when a stream is broken (e.g. too many packets are missing)."""


class RTPUnhandledPayload(RTPException, NotImplementedError):
    """Exception raised when a payload type is not handled by the library."""


class RTPUnsupportedCodec(RTPException, NotImplementedError):
    """Exception raised when a codec is not supported by this library."""


class VoIPException(SibilantException):
    """Base class for all exceptions raised by the VoIP module."""


class VoIPPhoneException(VoIPException):
    """Base class for all exceptions raised by the VoIPPhone class."""


class VoIPCallException(VoIPException):
    """Base class for all exceptions raised by the VoIPCall class."""


class VoIPCallTimeoutError(VoIPCallException, TimeoutError):
    """Raised when a VoIP call method times out."""



"""Helpers, utilities, and other miscellaneous functions and classes."""

from __future__ import annotations

import enum
import functools
import ipaddress
import logging
import re
import socket
import sys
import time
import types
import urllib.request
from abc import ABC, abstractmethod
from collections import OrderedDict
from collections.abc import (
    Callable,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    MutableSequence,
)
from dataclasses import dataclass as _dtcls, is_dataclass
from inspect import isabstract
from re import Pattern
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Protocol,
    TypeAlias,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
    overload,
    runtime_checkable,
)

import numpy as np
from typing_extensions import Self, dataclass_transform

from .constants import PUBLIC_IP_RESOLVERS


if TYPE_CHECKING:
    from _typeshed import SupportsKeysAndGetItem
    from numpy.typing import NDArray


_logger = logging.getLogger(__name__)


_dT = TypeVar("_dT")


@functools.wraps(_dtcls)
@dataclass_transform()
def slots_dataclass(*args: Any, **kwargs: Any) -> Callable[[_dT], _dT]:
    """Wrapper for dataclass decorator that adds slots if supported (py3.10+)."""
    # TODO: restore slots=True default once https://github.com/python/cpython/issues/91126 is fixed
    kwargs.setdefault("slots", True)
    return cast("Callable[[_dT], _dT]", _dtcls(*args, **kwargs))


@runtime_checkable
class SupportsStr(Protocol):
    """Protocol for objects that support str() conversion."""

    @abstractmethod
    def __str__(self) -> str: ...


class FieldsEnumDatatype:
    """Base class for datatypes wrapped in FieldsEnums."""

    @property
    def enum_value(self) -> Any:
        """The value to use for the enum member."""
        raise NotImplementedError(
            "Must be overridden by getting the value from the field"
        )

    def __str__(self) -> str:
        return str(self.enum_value)


class ValueMatchEnum(FieldsEnumDatatype, enum.Enum):
    """Mixin to match enum members by a specified property value."""

    _enum_value_map_: ClassVar[dict[Hashable, Self]]

    @classmethod
    def match_value(cls, value: Any) -> Self:
        """Match a value to a member."""
        if not hasattr(cls, "_enum_value_map_"):
            cls._enum_value_map_ = {
                cast("Self", member).enum_value: cast("Self", member)
                for member in cls._member_map_.values()
            }
        if value in cls._member_map_:
            return cast("Self", cls._member_map_[value])
        if value in cls._enum_value_map_:  # type: ignore[operator]
            return cast("Self", cls._enum_value_map_[value])  # type: ignore[index]
        raise ValueError(f"Not a valid {cls.__qualname__} value: {value}")


# noinspection PyTypeChecker
class FieldsEnum(ValueMatchEnum):
    """
    Custom enum class that's tied to a dataclass type and wraps its objects as members
    and proxies their attributes.
    """

    __wrapped_type__: ClassVar[type[FieldsEnumDatatype]]
    __allow_unknown__: ClassVar[bool] = False
    __unknown_member_name__: ClassVar[str] = "UNKNOWN"

    _wrapped_value_: Any

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        cls_name = cls.__name__
        if getattr(cls, "__wrapped_type__", None) is None:
            raise TypeError(f"{cls_name} must define __wrapped_type__")

    @classmethod
    def _get_enum_value(cls, objs: Any) -> Any:
        exc: BaseException | None = None
        for src in objs:
            try:
                return src.enum_value
            except (NotImplementedError, AttributeError) as exc:  # noqa: PERF203
                exc = exc  # noqa: PLW0127
        assert exc is not None
        raise exc

    def __new__(cls, value: Any) -> Self:  # noqa: D102
        if not isinstance(value, cls.__wrapped_type__) and isinstance(value, tuple):
            value = cls.__wrapped_type__(*value)
        if not isinstance(value, cls.__wrapped_type__):
            raise TypeError(
                f"Expected subclass of {cls.__wrapped_type__.__name__}, got {type(value)}"
            )
        if not is_dataclass(value):
            raise TypeError(f"Expected dataclass, got {type(value)}")
        assert isinstance(value, cls.__wrapped_type__)

        obj = object.__new__(cls)
        obj._wrapped_value_ = value

        enum_value = cls._get_enum_value((obj, value))
        obj._value_ = enum_value
        return obj

    __new_member__: ClassVar[Callable[[type[Self], Any], Self]]

    @classmethod
    def _missing_(cls, value: Any) -> FieldsEnum | None:
        if isinstance(value, cls.__wrapped_type__):
            try:
                return cls(cls._get_enum_value((value,)))
            except (ValueError, TypeError):
                if not cls.__allow_unknown__:
                    raise

        if not cls.__allow_unknown__:
            return None

        obj = cast("FieldsEnum", cls.__new_member__(cls, value))  # type: ignore[arg-type]
        obj._name_ = cls.__unknown_member_name__
        return obj

    def __getattr__(self, name: str) -> Any:
        return getattr(self._wrapped_value_, name)

    def __str__(self) -> str:
        return str(self._get_enum_value((self, self._wrapped_value_)))


_AUTO = types.new_class(
    "AUTO", bases=(FieldsEnumDatatype,)
)  # sentinel for AutoFieldsEnum wrapped type


# noinspection PyAbstractClass
# custom enum class that's tied to a dataclass and mirrors its fields on getattr
class AutoFieldsEnum(FieldsEnum):
    """Enum class that mirrors the fields on a dataclass."""

    __wrapped_type__ = _AUTO

    _dtcls_value_: Any

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Dynamically generate a dataclass from the enum definition, frozen, with slots,
        from the __annotations__ of this class.
        """
        dtcls_candidates = [c for c in cls.mro() if c is not cls and is_dataclass(c)]
        if not dtcls_candidates:
            raise TypeError(f"{cls.__qualname__} must inherit from a dataclass!")
        dtcls = dtcls_candidates[0]
        cls.__wrapped_type__ = dtcls

        super().__init_subclass__(**kwargs)

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:  # noqa: D102
        dtcls_value = cls.__wrapped_type__(*args, **kwargs)
        # yes, enum metaclasses make a mess of this
        obj = cast("Self", FieldsEnum.__new_member__(cls, dtcls_value))
        obj._dtcls_value_ = dtcls_value
        return obj


if sys.version_info >= (3, 11):
    DataclassEnum = ValueMatchEnum
else:
    DataclassEnum = AutoFieldsEnum


_T = TypeVar("_T")


# copied from requests.structures
class CaseInsensitiveDict(MutableMapping[str, _T]):
    """
    A case-insensitive ``dict``-like object.

    Implements all methods and operations of
    ``MutableMapping`` as well as dict's ``copy``. Also
    provides ``lower_items``.

    All keys are expected to be strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, ``items()``, ``iterkeys()``, and ``iteritems()``
    will contain case-sensitive keys. However, querying and contains
    testing is case insensitive::

        cid = CaseInsensitiveDict()
        cid['Accept'] = 'application/json'
        cid['aCCEPT'] == 'application/json'  # True
        list(cid) == ['Accept']  # True

    For example, ``headers['content-encoding']`` will return the
    value of a ``'Content-Encoding'`` response header, regardless
    of how the header name was originally stored.

    If the constructor, ``.update``, or equality comparison
    operations are given keys that have equal ``.lower()``s, the
    behavior is undefined.
    """

    def __init__(
        self,
        data: SupportsKeysAndGetItem[str, _T] | Iterable[tuple[str, _T]] | None = None,
        **kwargs: _T,
    ) -> None:
        self._store: OrderedDict[str, tuple[str, _T]] = OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key: str, value: _T) -> None:
        # Use the lowercased key for lookups, but store the actual
        # key alongside the value.
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key: str) -> _T:
        return self._store[key.lower()][1]

    def __delitem__(self, key: str) -> None:
        del self._store[key.lower()]

    def __iter__(self) -> Iterator[str]:
        return (casedkey for casedkey, mappedvalue in self._store.values())

    def __len__(self) -> int:
        return len(self._store)

    def lower_items(self) -> Iterator[tuple[str, _T]]:
        """Like iteritems(), but with all lowercase keys."""
        return ((lowerkey, keyval[1]) for (lowerkey, keyval) in self._store.items())

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    # Copy is required
    def copy(self) -> CaseInsensitiveDict[_T]:
        """Return a shallow copy of the instance."""
        return CaseInsensitiveDict(self._store.values())

    def __repr__(self) -> str:
        # noinspection PyTypeChecker
        return str(dict(self.items()))


def try_unpack_optional_type(typ_: Any) -> Any:
    """
    Unpack a type annotation that is Optional, or Union with None and a single other type.

    :param typ_: The type annotation
    :return: The original type wrapped in Optional, or the input argument if it's not Optional.
    """
    args = get_args(typ_)
    origin = get_origin(typ_)
    if (origin is Union or origin is getattr(types, "UnionType", Union)) and (
        args is not None and len(args) == 2 and type(None) in args
    ):
        return next(a for a in args if a is not type(None))
    return typ_


class _DefaultType:
    """Comparable and hashable sentinel for DEFAULT values."""

    # TODO: ensure singleton

    def __eq__(self, other: object) -> bool:
        return isinstance(other, _DefaultType)

    def __hash__(self) -> int:
        return hash((self.__class__, id(self)))

    def __repr__(self) -> str:
        return "DEFAULT"


DEFAULT = _DefaultType()
DefaultType: TypeAlias = _DefaultType


_ID = TypeVar("_ID")
_RT = TypeVar("_RT", bound="Registry")


class Registry(ABC, Generic[_ID, _RT]):
    """
    Abstract base class for registries of subclasses of a given class.

    Subclasses of this class can be initialized as registries, which can be then used
    to register subclasses of the given class through a specific class attribute.
    In the class declaration, some additional keyword arguments must be specified
    to properly initialize the registry:

    :param registry: whether the class is a registry or not.
    :param registry_attr: the name of the class attribute to use as the registry key.
    :param registry_attr_label: the label to use for the registry key in the generated
        methods and constructors.
    :param registry_attr_inheritable: whether the registry attribute is inheritable
        or must instead be always defined in the subclass body.

    Subclasses of "registry" classes are automatically registered in the registry
    using the value of the defined registry attribute as registry key, and can be
    retrieved through the generated :meth:`get_class_for_<attr_label>` method,
    or instantiated through the generated :meth:`for_<attr_label>` constructor.

    Abstract subclasses are not registered, only concrete ones are.
    """

    __registry__: MutableMapping[_ID, type[_RT]]
    __registry_attr_name__: str
    __registry_root__: type[Registry]

    @classmethod
    def is_abstract(cls) -> bool:
        """
        Check if the class is actually defined as abstract
        (i.e. has ABC in its bases, or abstract methods).
        """
        return isabstract(cls) or ABC in cls.__bases__

    # pylint: disable=arguments-differ
    def __init_subclass__(
        cls,
        *,
        registry: bool = False,
        registry_attr: str | None = None,
        registry_attr_label: str | None = None,
        registry_attr_inheritable: bool = True,
        **kwargs: Any,
    ):
        super().__init_subclass__(**kwargs)

        # Create a new registry
        if registry:
            if not registry_attr:
                raise AttributeError(
                    f"No attr_name specified for registry class {cls.__name__}"
                )
            if registry_attr_label is None:
                registry_attr_label = registry_attr

            cls.__registry__ = {}
            cls.__registry_attr_name__ = registry_attr
            cls.__registry_root__ = cls
            trimmed_attr_label: str = registry_attr_label.strip("_")
            setattr(
                cls,
                f"get_class_for_{trimmed_attr_label}",
                cls.__registry_get_class_for__,
            )
            setattr(cls, f"for_{trimmed_attr_label}", cls.__registry_new_for__)

        # Check registry subclass
        else:
            registry_id: str | None
            try:
                if registry_attr_inheritable:
                    registry_id = getattr(cls, cls.__registry_attr_name__)
                else:
                    registry_id = cls.__dict__.get(cls.__registry_attr_name__)
            except AttributeError:
                registry_id = None
            if registry_id is None:
                if cls.is_abstract():
                    return
                raise ValueError(
                    f"Cannot register {cls.__name__} in {cls.__registry_root__.__name__}, "
                    f"no {cls.__registry_attr_name__} defined in the class body"
                )

            conflict_cls: type[_RT] | None = cls.__registry__.get(registry_id)
            if conflict_cls is not None:
                cls_fullname = (cls.__module__, cls.__qualname__)
                conflict_fullname = (conflict_cls.__module__, conflict_cls.__qualname__)
                if cls_fullname != conflict_fullname:  # not reinit
                    raise NameError(
                        f"More than one {cls.__registry_root__.__name__} subclass with "
                        f'the same {cls.__registry_attr_name__} "{registry_id}" defined: '
                        f"{conflict_cls.__name__} and {cls.__name__}"
                    )
            cls.__registry__[registry_id] = cls

    @classmethod
    def get_registry(cls) -> types.MappingProxyType[_ID, type[_RT]]:
        """Get a read-only view of the registry mapping."""
        return types.MappingProxyType(cls.__registry__)

    @classmethod
    def __registry_get_class_for__(cls, registry_id: _ID) -> type[_RT]:
        registered_cls: type[_RT] | None = cls.__registry__.get(registry_id)
        if registered_cls is None:
            raise KeyError(
                f"No registered {cls.__registry_root__.__name__} subclass found "
                f'for {cls.__registry_attr_name__} == "{registry_id}"'
            )
        return registered_cls

    @classmethod
    def __registry_new_for__(cls, registry_id: _ID, *args: Any, **kwargs: Any) -> _RT:
        registered_cls: type[_RT] = cls.__registry_get_class_for__(registry_id)
        # noinspection PyArgumentList
        return cast("_RT", registered_cls(*args, **kwargs))


_sT_contra = TypeVar("_sT_contra", str, bytes, contravariant=True)
_sT_co = TypeVar("_sT_co", str, bytes, covariant=True)


@runtime_checkable
class ParseableAny(Protocol[_sT_contra]):
    """Generic protocol for parseable objects from AnyStr."""

    @classmethod
    def parse(cls, raw_value: _sT_contra) -> Self:
        """Parse a string value into an instance of this class."""


@runtime_checkable
class SerializableAny(Protocol[_sT_co]):
    """Generic protocol for objects serializable to AnyStr."""

    def serialize(self) -> _sT_co:
        """Serialize the object to a string."""


@runtime_checkable
class ParseableSerializableAny(
    ParseableAny[_sT_contra], SerializableAny[_sT_co], Protocol[_sT_contra, _sT_co]
):
    """Generic protocol for objects that are both parseable and serializable to AnyStr."""


@runtime_checkable
class Parseable(ParseableAny[str], Protocol):
    """Generic protocol for parseable objects from str."""


@runtime_checkable
class Serializable(SerializableAny[str], Protocol):
    """Generic protocol for objects serializable to str."""


@runtime_checkable
class ParseableSerializable(Parseable, Serializable, Protocol):
    """Generic protocol for objects that are both parseable and serializable to str."""


@runtime_checkable
class ParseableRaw(ParseableAny[bytes], Protocol):
    """Generic protocol for parseable objects from bytes."""

    @classmethod
    def parse(cls, raw_value: bytes) -> Self:
        """Parse a bytes value into an instance of this class."""


@runtime_checkable
class SerializableRaw(SerializableAny[bytes], Protocol):
    """Generic protocol for objects serializable to bytes."""

    def serialize(self) -> bytes:
        """Serialize the object to bytes."""


@runtime_checkable
class ParseableSerializableRaw(ParseableRaw, SerializableRaw, Protocol):
    """Generic protocol for objects that are both parseable and serializable to bytes."""


@runtime_checkable
class FieldsParser(Protocol):
    """Generic protocol for objects that can parse a string values into separate fields."""

    # FIXME: rename to e.g. parse_fields (also in subclasses, like SDP, SIP Headers, etc.)
    @classmethod
    def parse_raw_value(cls, raw_value: str) -> dict[str, Any]:
        """Parse a string value into a mapping of fields values."""


@runtime_checkable
class FieldsParserSerializer(FieldsParser, Serializable, Protocol):
    """
    Generic protocol for objects that can parse a string values into separate fields
    and serialize them back into a string.
    """


@slots_dataclass
class StrValueMixin(FieldsParserSerializer):
    """Mixin for dataclasses that have a single string field and can be parsed/serialized."""

    value: str

    @classmethod
    def parse_raw_value(cls, raw_value: str) -> dict[str, Any]:  # noqa: D102
        return dict(value=raw_value)

    def serialize(self) -> str:  # noqa: D102
        return self.value


@slots_dataclass
class OptionalStrValueMixin(FieldsParserSerializer):
    """Mixin like :class:`StrValueMixin`, but that also accepts empty values (`None`)."""

    value: str | None

    @classmethod
    def parse_raw_value(cls, raw_value: str) -> dict[str, Any]:  # noqa: D102
        return dict(value=raw_value)

    def serialize(self) -> str:  # noqa: D102
        return self.value or ""


@slots_dataclass
class IntValueMixin(FieldsParserSerializer):
    """Mixin for dataclasses that have a single integer field and can be parsed/serialized."""

    value: int

    @classmethod
    def parse_raw_value(cls, raw_value: str) -> dict[str, Any]:  # noqa: D102
        return dict(value=int(raw_value))

    def serialize(self) -> str:  # noqa: D102
        return str(self.value)

    def __int__(self) -> int:
        return self.value


_ST = TypeVar("_ST", bound=SupportsStr | ParseableSerializable)


@slots_dataclass
class ListValueMixin(MutableSequence, FieldsParserSerializer, Generic[_ST]):
    """
    Mixin for dataclasses that have a list of values and can be parsed/serialized.

    Also provides a list-like interface to access the values.
    """

    _values_type: ClassVar[type[_ST]]  # type: ignore[misc]
    _separator: ClassVar[str] = ", "
    _splitter: ClassVar[str | Pattern[str] | None] = re.compile(r"\s*,\s*")

    values: list[_ST]
    raw_value: str = ""

    def __post_init__(self) -> None:
        if not self.raw_value:
            self.raw_value = self._serialize()

    @classmethod
    def parse_raw_value(cls, raw_value: str) -> dict[str, Any]:  # noqa: D102
        str_value = raw_value.strip()
        splitter = cls._splitter or cls._separator
        str_values: list[str]
        if not str_value:
            str_values = []
        elif isinstance(splitter, str):
            str_values = str_value.split(splitter)
        elif isinstance(splitter, Pattern):
            str_values = splitter.split(str_value)
        else:
            raise TypeError(f"Invalid splitter for {cls.__name__}: {splitter!r}")
        vcls = cls._values_type
        values: list[_ST] = [
            cast(
                "_ST", vcls.parse(value) if issubclass(vcls, Parseable) else vcls(value)
            )
            for value in str_values
        ]
        return dict(values=values, raw_value=raw_value)

    def _serialized_values(self) -> list[str]:
        return [
            value.serialize() if isinstance(value, Serializable) else str(value)
            for value in self.values
        ]

    def _serialize(self) -> str:
        return self._separator.join(self._serialized_values())

    def serialize(self) -> str:  # noqa: D102
        return self.raw_value or self._serialize()

    def insert(self, index: int, value: _ST) -> None:  # noqa: D102
        self.values.insert(index, value)

    @overload
    def __getitem__(self, index: int) -> _ST: ...

    @overload
    def __getitem__(self, index: slice) -> MutableSequence[_ST]: ...

    def __getitem__(self, index: int | slice) -> _ST | MutableSequence[_ST]:
        return self.values[index]

    @overload
    def __setitem__(self, index: int, value: _ST) -> None: ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[_ST]) -> None: ...

    def __setitem__(self, index: int | slice, value: _ST | Iterable[_ST]) -> None:
        self.values[index] = value  # type: ignore[index,assignment]

    def __delitem__(self, index: int | slice) -> None:
        del self.values[index]

    def __len__(self) -> int:
        return len(self.values)


_rV_co = TypeVar("_rV_co", covariant=True)


class _TimeCachedCallable(Protocol[_rV_co]):
    _time_cache_expiry: float

    cache_info: Callable[[], Any]
    cache_clear: Callable[[], None]

    def __call__(self, *args: Any, **kwargs: Any) -> _rV_co: ...


def time_cache(
    expiry: float, *, maxsize: int = 1, typed: bool = False
) -> Callable[[Callable[..., _rV_co]], Callable[..., _rV_co]]:
    """Simple time / expiration cache decorator, implmented atop functools.lru_cache."""

    def decorator(func: Callable[..., _rV_co]) -> Callable[..., _rV_co]:
        def _wrapper(*args: Any, **kwargs: Any) -> _rV_co:
            return func(*args, **kwargs)

        wrapper = cast(
            "_TimeCachedCallable[_rV_co]",
            functools.lru_cache(maxsize=maxsize, typed=typed)(_wrapper),
        )

        @functools.wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> _rV_co:
            now = time.monotonic()
            if now > wrapper._time_cache_expiry:
                wrapper.cache_clear()
                wrapper._time_cache_expiry = now + expiry
            return wrapper(*args, **kwargs)

        wrapper._time_cache_expiry = 0.0
        return wrapped

    return decorator


def is_socket_bound(sock: socket.socket) -> bool:
    """Check if a socket is bound to a local address."""
    if sock.family not in {socket.AF_INET, socket.AF_INET6}:
        raise TypeError(f"Unsupported socket family: {sock.family}")
    try:
        bind_addr = sock.getsockname()
        return bool(bind_addr[1])  # if port is 0 then it's not bound
    except OSError:
        return False


@time_cache(expiry=60.0)
def get_public_ip() -> str:
    """Get the public IP address of the current machine."""

    def try_resolver(
        url: str, extract_fn: Callable[[str], str | None] | None = None
    ) -> str:
        body = urllib.request.urlopen(url, timeout=10).read().decode("utf8")  # noqa: S310
        if extract_fn:
            body = extract_fn(body)
        if not body:
            raise ValueError(f"Could not extract public IP address from {url} response")
        ip = ipaddress.ip_address(body.strip())
        return str(ip)

    for url, extract_fn in PUBLIC_IP_RESOLVERS:
        try:
            return try_resolver(url, extract_fn)
        except Exception as e:  # noqa: BLE001, PERF203
            _logger.warning(f"Failed to get public IP address from: {url!r}: {e}")
    raise RuntimeError("Could not resolve public IP address")


def get_local_ip_for_dest(host: str) -> str:
    """Get the IP address of the current machine relative to the given host on the local network."""
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect((host, 0))
        return cast("str", s.getsockname()[0])


def get_external_ip_for_dest(host: str) -> str:
    """Get the IP address of the current machine relative to the given host."""
    # resolve host if it's not an IP address
    host_ip = socket.gethostbyname(host)
    is_private = ipaddress.ip_address(host_ip).is_private
    if is_private:
        return get_local_ip_for_dest(host_ip)
    else:
        return get_public_ip()


@overload
def db_to_amplitude(db: float, *, ref: float = 1.0) -> float: ...


@overload
def db_to_amplitude(
    db: NDArray[np.floating], *, ref: float = 1.0
) -> NDArray[np.floating]: ...


@overload
def db_to_amplitude(
    db: float | NDArray[np.floating], *, ref: float = 1.0
) -> float | NDArray[np.floating]: ...


def db_to_amplitude(
    db: float | NDArray[np.floating], *, ref: float = 1.0
) -> float | NDArray[np.floating]:
    """Convert dB-scaled values to amplitude."""
    return ((ref**2) * np.power(10.0, db * 0.1)) ** 0.5



"""Common base classes for SDP sections, fields and attributes."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import deque
from dataclasses import InitVar, dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    cast,
    get_origin,
    get_type_hints,
)

from typing_extensions import Self, override

from sibilant.exceptions import SDPParseError, SDPUnknownFieldError
from sibilant.helpers import (
    DEFAULT,
    DefaultType,
    FieldsParser,
    OptionalStrValueMixin,
    ParseableSerializable,
    ParseableSerializableRaw,
    Registry,
    StrValueMixin,
    try_unpack_optional_type,
)


if TYPE_CHECKING:
    from collections.abc import MutableMapping


__all__ = [
    "FlagAttribute",
    "InactiveFlag",
    "MediaFlowAttribute",
    "RecvOnlyFlag",
    "SDPAttribute",
    "SDPAttributeField",
    "SDPBandwidthField",
    "SDPConnectionField",
    "SDPEncryptionField",
    "SDPField",
    "SDPInformationField",
    "SDPSection",
    "SendOnlyFlag",
    "SendRecvFlag",
    "UnknownAttribute",
]


@dataclass
class SDPField(Registry[str, "SDPField"], ParseableSerializable, ABC):
    """Abstract base dataclass for SDP fields."""

    _type: ClassVar[str]
    _description: ClassVar[str]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        # make sure type is set and register the class into known types
        if ABC not in cls.__bases__ and not getattr(cls, "_description", None):
            raise ValueError(f"SDPField class {cls} must have a _description attribute")

    @property
    def type(self) -> str:
        """The type of the field."""
        return self._type

    @classmethod
    def parse(cls, raw_data: str) -> Self:  # noqa: D102
        field_type, raw_value = raw_data.strip().split("=", 1)

        try:
            field_cls = cls.__registry_get_class_for__(field_type)
        except KeyError:
            raise SDPUnknownFieldError(f"Unknown SDP field type {field_type}")

        return cast(
            "Self", field_cls.from_raw_value(field_type=field_type, raw_value=raw_value)
        )

    @classmethod
    def from_raw_value(cls, field_type: str, raw_value: str) -> Self:
        """
        Parse the raw value of the field into a field object.

        :param field_type: the field type
        :param raw_value: the raw value of the field
        :return: the field object.
        """
        if isinstance(cls, FieldsParser):
            return cls(**cls.parse_raw_value(raw_value))
        raise NotImplementedError

    @abstractmethod
    def serialize(self) -> str:
        """
        Serialize the field value to a string.

        :return: The serialized field value string.
        """

    def __str__(self) -> str:
        return f"{self.type}={self.serialize()}"


@dataclass
class SDPAttribute(
    Registry[str | DefaultType, "SDPAttribute"], ParseableSerializable, ABC
):
    """Abstract base dataclass for SDP attributes."""

    _name: ClassVar[str | DefaultType]
    _is_flag: ClassVar[bool | None] = None

    @property
    def name(self) -> str:
        """The name of the attribute."""
        if self._name is DEFAULT:
            raise SyntaxError(
                f"Class {self.__class__} must override name() property when using _name = DEFAULT"
            )
        assert isinstance(self._name, str)
        return self._name

    @property
    def is_flag(self) -> bool:
        """Whether the attribute is a flag or not."""
        return bool(self._is_flag)

    @classmethod
    def parse(cls, raw_data: str) -> Self:  # noqa: D102
        name: str
        raw_value: str | None
        name, raw_value = (
            raw_data.split(":", 1) if ":" in raw_data else (raw_data, None)  # type: ignore[assignment]
        )

        if cls._is_flag is not None:
            if cls._is_flag and raw_value is not None:
                raise SDPParseError(
                    f"Attribute {name} is a flag, but got a value: {raw_data}"
                )
            if not cls._is_flag and raw_value is None:
                raise SDPParseError(
                    f"Attribute {name} is not a flag, but got no value: {raw_data}"
                )

        registry_name: str = name.lower()
        is_known_attribute: bool = registry_name in cls.__registry__
        if not is_known_attribute and DEFAULT not in cls.__registry__:
            raise TypeError(
                f"Unknown SDP attribute {name}, and no default attribute class is defined"
            )
        attr_cls: type[SDPAttribute] = cls.__registry_get_class_for__(
            registry_name if is_known_attribute else DEFAULT
        )

        return cast("Self", attr_cls.from_raw_value(name, raw_value))

    @classmethod
    @abstractmethod
    def from_raw_value(cls, name: str, raw_value: str | None) -> Self:
        """
        Parse a raw value into an instance of this attribute class.

        :param name: the name of the attribute parsed from raw data
        :param raw_value: the raw value of the attribute parsed from raw data
        :return:
        """

    @abstractmethod
    def serialize(self) -> str:
        """Serialize the attribute value to a string."""

    def __str__(self) -> str:
        """Serialize the whole attribute to a string."""
        return f"{self.name}:{self.serialize()}" if not self.is_flag else self.name


@dataclass
class FlagAttribute(SDPAttribute, ABC):
    """Abstract base dataclass for SDP flag attributes."""

    _is_flag: ClassVar[bool] = True

    @classmethod
    def from_raw_value(cls, name: str, raw_value: str | None) -> Self:  # noqa: D102
        return cls()

    def serialize(self) -> str:  # noqa: D102
        raise ValueError("Flag attributes have no value to serialize")


@dataclass
class ValueAttribute(SDPAttribute, ABC):
    """Abstract base dataclass for SDP value attributes."""

    value: Any

    @classmethod
    def from_raw_value(cls, name: str, raw_value: str | None) -> Self:
        return cls(value=raw_value)


@dataclass
class UnknownAttribute(OptionalStrValueMixin, SDPAttribute, ABC):
    """Abstract base dataclass for parsing unsupported SDP attributes."""

    _name = DEFAULT

    attribute: str

    @property
    def name(self) -> str:
        """The name of the attribute."""
        return self.attribute

    @classmethod
    def from_raw_value(cls, name: str, raw_value: str | None) -> Self:  # noqa: D102
        return cls(attribute=name, value=raw_value)


class MediaFlowAttribute(FlagAttribute, ABC):
    """Abstract base dataclass for SDP media flow attributes, defined in :rfc:`8866#section-6.7`."""

    _is_flag = True


@dataclass
class RecvOnlyFlag(MediaFlowAttribute, ABC):
    """SDP media flow attribute for recvonly, defined in :rfc:`8866#section-6.7.1`."""

    _name = "recvonly"


@dataclass
class SendRecvFlag(MediaFlowAttribute, ABC):
    """SDP media flow attribute for sendrecv, defined in :rfc:`8866#section-6.7.2`."""

    _name = "sendrecv"


@dataclass
class SendOnlyFlag(MediaFlowAttribute, ABC):
    """SDP media flow attribute for sendonly, defined in :rfc:`8866#section-6.7.3`."""

    _name = "sendonly"


@dataclass
class InactiveFlag(MediaFlowAttribute, ABC):
    """SDP media flow attribute for inactive, defined in :rfc:`8866#section-6.7.4`."""

    _name = "inactive"


@dataclass
class SDPInformationField(StrValueMixin, SDPField, ABC):
    """
    SDP session information field, defined in :rfc:`8866#section-5.4`.

    Spec::
        i=<session description>
    """

    _type = "i"

    @property
    def session_description(self) -> str:
        """The session description."""
        return self.value


@dataclass
class SDPConnectionField(SDPField, ABC):
    """
    SDP session connection field, defined in :rfc:`8866#section-5.7`.

    Spec::
        c=<nettype> <addrtype> <connection-address>
    """

    _type = "c"

    nettype: str
    addrtype: str
    address: str
    ttl: int | None = None
    number_of_addresses: int | None = None

    @property
    def connection_address(self) -> str:
        """The connection address as string, with optional TTL and number of addresses."""
        parts = [self.address]
        if self.ttl is not None:
            parts.append(str(self.ttl))
        if self.number_of_addresses is not None:
            parts.append(str(self.number_of_addresses))
        return "/".join(parts)

    @classmethod
    @override
    def from_raw_value(cls, field_type: str, raw_value: str) -> Self:
        nettype, addrtype, connection_address = raw_value.split(" ")
        # parse the connection address. Keep in mind IPv6 don't have TTL,
        # so the second part might be the number of addresses
        address, *rest = connection_address.split("/")
        ttl = number_of_addresses = None
        if rest:
            if addrtype == "IP6":
                if len(rest) >= 1:
                    raise SDPParseError(
                        f"Invalid connection address {connection_address}"
                    )
                number_of_addresses = int(rest[0])
            else:
                if len(rest) >= 2:
                    raise SDPParseError(
                        f"Invalid connection address {connection_address}"
                    )
                ttl = int(rest[0])
                if len(rest) == 2:
                    number_of_addresses = int(rest[1])
        return cls(
            nettype=nettype,
            addrtype=addrtype,
            address=address,
            ttl=ttl,
            number_of_addresses=number_of_addresses,
        )

    def serialize(self) -> str:  # noqa: D102
        return " ".join((self.nettype, self.addrtype, self.connection_address))  # noqa: FLY002


@dataclass
class SDPBandwidthField(SDPField, ABC):
    """
    SDP session bandwidth field, defined in :rfc:`8866#section-5.8`.

    Spec::
        b=<bwtype>:<bandwidth>
    """

    _type = "b"

    bwtype: str
    bandwidth: int

    @classmethod
    @override
    def from_raw_value(cls, field_type: str, raw_value: str) -> Self:
        bwtype, bandwidth = raw_value.split(":")
        return cls(bwtype=bwtype, bandwidth=int(bandwidth))

    def serialize(self) -> str:  # noqa: D102
        return f"{self.bwtype}:{self.bandwidth}"


@dataclass
class SDPEncryptionField(SDPField, ABC):
    """
    SDP session encryption field, defined in :rfc:`8866#section-5.12`.

    Spec::
        k=<method>
        k=<method>:<encryption key>
    """

    _type = "k"

    method: str
    key: str | None = None

    @classmethod
    @override
    def from_raw_value(cls, field_type: str, raw_value: str) -> Self:
        method: str
        key: str | None
        method, key = raw_value.split(":") if ":" in raw_value else (raw_value, None)  # type: ignore[assignment]
        return cls(method=method, key=key)

    def serialize(self) -> str:  # noqa: D102
        return f"{self.method}:{self.key}" if self.key else self.method


@dataclass
class SDPAttributeField(SDPField, ABC):
    """Abstract base dataclass for SDP attribute fields."""

    _type = "a"
    _attribute_cls: ClassVar[type[SDPAttribute]]

    attribute: SDPAttribute

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        if not hasattr(cls, "_attribute_cls"):
            raise TypeError(
                f"Attribute field class {cls} must have an attribute class defined"
            )

    @property
    def name(self) -> str:
        """The name of the attribute."""
        return self.attribute.name

    @property
    def is_flag(self) -> bool:
        """Whether the attribute is a flag or not."""
        return self.attribute.is_flag

    @classmethod
    @override
    def from_raw_value(cls, field_type: str, raw_value: str) -> Self:
        return cls(attribute=cls._attribute_cls.parse(raw_value))

    def serialize(self) -> str:  # noqa: D102
        return str(self.attribute)


@dataclass
class SDPSection(ParseableSerializableRaw, ABC):
    """Abstract base dataclass for SDP sections."""

    _fields_base: ClassVar[type[SDPField]]
    _start_field: ClassVar[type[SDPField]]

    # mapping of {sdptype: (field_name, field_type, wrapped_type), ...}
    _sdp_fields_map: ClassVar[dict[str, tuple[str, Any, type]]]
    _subsections_map: ClassVar[dict[str, type[SDPSection]]]

    @classmethod
    def _reveal_wrapped_type(cls, field_type: Any) -> type:
        if isinstance(field_type, str) and field_type in globals():
            field_type = globals()[field_type]
        if get_origin(field_type) in {list, list}:
            field_type = field_type.__args__[0]
        field_type = try_unpack_optional_type(field_type)
        if isinstance(field_type, type):
            return field_type
        raise TypeError(f"Unknown field type {field_type}")

    @classmethod
    def _init_fields_map(cls) -> None:
        cls._sdp_fields_map = {}
        cls._subsections_map = {}
        # check annotations
        for field_name, field_type in get_type_hints(cls).items():
            # skip ClassVar and InitVar fields
            if get_origin(field_type) in {ClassVar, InitVar}:
                continue
            wrapped_type: type = cls._reveal_wrapped_type(field_type)
            sdp_type: str | None = None
            if issubclass(wrapped_type, SDPField):
                sdp_type = wrapped_type._type  # noqa: SLF001
            elif issubclass(wrapped_type, SDPSection):
                # noinspection PyProtectedMember
                sdp_type = wrapped_type._start_field._type  # noqa: SLF001
                cls._subsections_map[sdp_type] = wrapped_type
            elif not issubclass(wrapped_type, SDPField):
                continue
            if sdp_type is None:
                raise TypeError(f"Unknown field type {wrapped_type}")
            cls._sdp_fields_map[sdp_type] = field_name, field_type, wrapped_type

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        if not hasattr(cls, "_fields_base"):
            raise TypeError(f"SDPSection subclass {cls} must define _fields_base")

        if not hasattr(cls, "_start_field"):
            raise TypeError(f"SDPSection subclass {cls} must define _start_field")

        cls._init_fields_map()

    @classmethod
    def _line_preprocess(cls, line: str, fields: MutableMapping[str, Any]) -> str:
        return line

    @classmethod
    def from_lines(cls, lines: deque[str], *, is_subsection: bool = False) -> Self:
        """
        Parse an SDP section from a list of lines.

        :param lines: the lines to parse.
        :param is_subsection: whether the section is a subsection of another section.
        :return: the parsed SDP section.
        """
        fields: dict[str, Any] = {}
        while lines:
            line = lines.popleft()
            value: SDPField | SDPSection | None = None
            sdp_type: str | None = None
            for sdp_type, subsection_type in cls._subsections_map.items():
                if line.startswith(sdp_type + "="):
                    lines.appendleft(line)
                    value = subsection_type.from_lines(lines, is_subsection=True)
                    break
            else:
                line = cls._line_preprocess(line, fields)
                try:
                    value = cls._fields_base.parse(line)  # could raise SDPParseError
                except SDPUnknownFieldError:
                    if is_subsection:
                        lines.appendleft(line)
                        break
                else:
                    assert isinstance(value, SDPField)
                    sdp_type = value.type

            assert sdp_type is not None
            assert value is not None
            field_name, field_type, _ = cls._sdp_fields_map[sdp_type]
            if get_origin(field_type) in {list, list}:
                fields.setdefault(field_name, []).append(value)
            else:
                if field_name in fields:
                    raise SDPParseError(f"Duplicate field {field_name}")
                fields[field_name] = value

        # noinspection PyArgumentList
        return cls(**fields)

    @classmethod
    def parse(cls, raw_value: bytes) -> Self:
        """Parse an SDP session from a string."""
        sdp_lines = raw_value.decode("utf-8").split("\r\n")

        to_process_lines: deque[str] = deque(
            stripped_line for line in sdp_lines if (stripped_line := line.strip())
        )
        return cls.from_lines(to_process_lines)

    def serialize(self) -> bytes:
        """Serialize the SDP section to a string."""
        return str(self).encode("utf-8")

    def __str__(self) -> str:
        """Serialize the SDP section to a string."""
        serialized_fields: list[str] = []
        for field_name, field_type, _ in self._sdp_fields_map.values():
            value = getattr(self, field_name)
            if value is None:
                continue
            if get_origin(field_type) in {list, list}:
                serialized_fields.extend(map(str, value))
            else:
                serialized_fields.append(str(value))

        return "\r\n".join(serialized_fields)



"""SDP session section and fields definitions and implementations."""

from __future__ import annotations

import logging
from abc import ABC
from dataclasses import dataclass, field as dataclass_field
from typing import TYPE_CHECKING, Any

from typing_extensions import Self, override

from sibilant.constants import SUPPORTED_SDP_VERSIONS
from sibilant.exceptions import SDPParseError
from sibilant.helpers import FieldsParserSerializer, StrValueMixin, slots_dataclass

from .common import (
    InactiveFlag,
    RecvOnlyFlag,
    SDPAttribute,
    SDPAttributeField,
    SDPBandwidthField,
    SDPConnectionField,
    SDPEncryptionField,
    SDPField,
    SDPInformationField,
    SDPSection,
    SendOnlyFlag,
    SendRecvFlag,
    UnknownAttribute,
)
from .media import SDPMedia, get_media_flow_type
from .time import SDPTime  # noqa


if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from sibilant.rtp import MediaFlowType


__all__ = [
    "InactiveSessionFlag",
    "RecvOnlySessionFlag",
    "SDPSession",
    "SDPSessionAttribute",
    "SDPSessionAttributeField",
    "SDPSessionBandwidth",
    "SDPSessionConnection",
    "SDPSessionEmail",
    "SDPSessionEncryption",
    "SDPSessionFields",
    "SDPSessionInformation",
    "SDPSessionName",
    "SDPSessionOrigin",
    "SDPSessionPhone",
    "SDPSessionTimezone",
    "SDPSessionTimezoneAdjustment",
    "SDPSessionURI",
    "SDPSessionVersion",
    "SendOnlySessionFlag",
    "SendRecvSessionFlag",
    "UnknownSessionAttribute",
]


_logger = logging.getLogger(__name__)


@dataclass
class SDPSessionFields(SDPField, ABC, registry=True, registry_attr="_type"):
    """Base class for SDP session description fields."""


@slots_dataclass
class SDPSessionVersion(StrValueMixin, SDPSessionFields):
    """
    SDP version field, defined in :rfc:`8866#section-5.1`.

    Spec::
        v=0.
    """

    _type = "v"
    _description = "protocol version"

    def __post_init__(self) -> None:
        if self.value not in SUPPORTED_SDP_VERSIONS:
            raise SDPParseError(f"Unsupported SDP version {self.value}")


@slots_dataclass
class SDPSessionOrigin(SDPSessionFields):
    """
    SDP origin field, defined in :rfc:`8866#section-5.2`.

    Spec::
        o=<username> <sess-id> <sess-version> <nettype> <addrtype> <unicast-address>
    """

    _type = "o"
    _description = "originator and session identifier"

    username: str
    sess_id: str
    sess_version: str
    nettype: str
    addrtype: str
    unicast_address: str

    @classmethod
    @override
    def from_raw_value(cls, field_type: str, raw_value: str) -> Self:
        (
            username,
            sess_id,
            sess_version,
            nettype,
            addrtype,
            unicast_address,
        ) = raw_value.split(" ")
        return cls(
            username=username,
            sess_id=sess_id,
            sess_version=sess_version,
            nettype=nettype,
            addrtype=addrtype,
            unicast_address=unicast_address,
        )

    def serialize(self) -> str:  # noqa: D102
        return " ".join((
            str(self.username),
            str(self.sess_id),
            str(self.sess_version),
            str(self.nettype),
            str(self.addrtype),
            str(self.unicast_address),
        ))


@slots_dataclass
class SDPSessionName(StrValueMixin, SDPSessionFields):
    """
    SDP session name field, defined in :rfc:`8866#section-5.3`.

    Spec::
        s=<session name>
    """

    _type = "s"
    _description = "session name"

    @property
    def session_name(self) -> str:
        """The session name."""
        return self.value


@slots_dataclass
class SDPSessionInformation(SDPInformationField, SDPSessionFields):
    """
    SDP session information field, defined in :rfc:`8866#section-5.4`.

    Spec::
        i=<session description>
    """

    _description = "session information"


@slots_dataclass
class SDPSessionURI(StrValueMixin, SDPSessionFields):
    """
    SDP session URI field, defined in :rfc:`8866#section-5.5`.

    Spec::
        u=<uri>
    """

    _type = "u"
    _description = "URI of description"

    @property
    def uri(self) -> str:
        """The URI."""
        return self.value


@slots_dataclass
class SDPSessionEmail(StrValueMixin, SDPSessionFields):
    """
    SDP session email field, defined in :rfc:`8866#section-5.6`.

    Spec::
        e=<email-address>
    """

    _type = "e"
    _description = "email address"

    @property
    def email_address(self) -> str:
        """The email address."""
        return self.value


@slots_dataclass
class SDPSessionPhone(StrValueMixin, SDPSessionFields):
    """
    SDP session phone field, defined in :rfc:`8866#section-5.6`.

    Spec::
        p=<phone-number>
    """

    _type = "p"
    _description = "phone number"

    @property
    def phone_number(self) -> str:
        """The phone number."""
        return self.value


@slots_dataclass
class SDPSessionConnection(SDPConnectionField, SDPSessionFields):
    """
    SDP session connection field, defined in :rfc:`8866#section-5.7`.

    Spec::
        c=<nettype> <addrtype> <connection-address>
    """

    _description = "connection information -- not required if included in all media"


@slots_dataclass
class SDPSessionBandwidth(SDPBandwidthField, SDPSessionFields):
    """
    SDP session bandwidth field, defined in :rfc:`8866#section-5.8`.

    Spec::
        b=<bwtype>:<bandwidth>
    """

    _description = "zero or more bandwidth information lines"


@slots_dataclass
class SDPSessionTimezoneAdjustment(FieldsParserSerializer):
    """
    SDP session timezone adjustments, as part of definition in :rfc:`8866#section-5.11`.

    Spec::
        <adjustment time> <offset>
    """

    adjustment_time: int
    offset: str

    @classmethod
    def from_raw_value(cls, raw_value: str) -> Self:  # noqa: D102
        return cls(**cls.parse_raw_value(raw_value))

    @classmethod
    def parse_raw_value(cls, raw_value: str) -> dict[str, Any]:  # noqa: D102
        adjustment_time, offset = raw_value.split(" ")
        return dict(adjustment_time=int(adjustment_time), offset=offset)

    def serialize(self) -> str:  # noqa: D102
        return f"{self.adjustment_time} {self.offset}"

    def __str__(self) -> str:
        return self.serialize()


@slots_dataclass
class SDPSessionTimezone(SDPSessionFields, FieldsParserSerializer):
    """
    SDP session timezone field, defined in :rfc:`8866#section-5.11`.

    Spec::
        z=<adjustment time> <offset> <adjustment time> <offset> ....
    """

    _type = "z"
    _description = "time zone adjustments"

    adjustments: list[SDPSessionTimezoneAdjustment]

    @classmethod
    def parse_raw_value(cls, raw_value: str) -> dict[str, Any]:  # noqa: D102
        split_values = raw_value.split(" ")
        if len(split_values) % 2 != 0:
            raise SDPParseError(
                f"Number of values in timezone field is not even (got {len(split_values)}): "
                f"{raw_value}"
            )
        adjustments = [
            SDPSessionTimezoneAdjustment.from_raw_value(f"{adjustment_time} {offset}")
            for adjustment_time, offset in zip(
                split_values[::2], split_values[1::2], strict=True
            )
        ]
        return dict(adjustments=adjustments)

    def serialize(self) -> str:  # noqa: D102
        return " ".join(str(adjustment) for adjustment in self.adjustments)


@slots_dataclass
class SDPSessionEncryption(SDPEncryptionField, SDPSessionFields):
    """
    SDP session encryption field, defined in :rfc:`8866#section-5.12`.

    Spec::
        k=<method>
        k=<method>:<encryption key>
    """

    _description = "encryption key"


@dataclass
class SDPSessionAttribute(SDPAttribute, ABC, registry=True, registry_attr="_name"):
    """Base class for SDP session attributes."""


@slots_dataclass
class UnknownSessionAttribute(UnknownAttribute, SDPSessionAttribute):
    """Catch-all class for unsupported SDP session attributes."""


@slots_dataclass
class RecvOnlySessionFlag(RecvOnlyFlag, SDPSessionAttribute):
    """
    SDP session attribute for recvonly media flow, defined in :rfc:`8866#section-6.7.1`.

    spec::
        recvonly
    """


@slots_dataclass
class SendRecvSessionFlag(SendRecvFlag, SDPSessionAttribute):
    """
    SDP session attribute for sendrecv media flow, defined in :rfc:`8866#section-6.7.2`.

    spec::
        sendrecv
    """


@slots_dataclass
class SendOnlySessionFlag(SendOnlyFlag, SDPSessionAttribute):
    """
    SDP session attribute for sendonly media flow, defined in :rfc:`8866#section-6.7.3`.

    spec::
        sendonly
    """


@slots_dataclass
class InactiveSessionFlag(InactiveFlag, SDPSessionAttribute):
    """
    SDP session attribute for inactive media flow, defined in :rfc:`8866#section-6.7.4`.

    spec::
        inactive
    """


@slots_dataclass
class SDPSessionAttributeField(SDPAttributeField, SDPSessionFields):
    """
    SDP session attribute field, defined in :rfc:`8866#section-5.13`.

    Spec::
        a=<attribute>
        a=<attribute>:<value>
    """

    _attribute_cls = SDPSessionAttribute

    _description = "zero or more session attribute lines"


# TODO: RFC 8866 requires that the order of fields is exactly as defined in the spec. Add tests.


@dataclass
class SDPSession(SDPSection):
    """SDP section for session description fields, defined in :rfc:`8866#section-5`."""

    _fields_base = SDPSessionFields
    _start_field = SDPSessionVersion

    version: SDPSessionVersion
    origin: SDPSessionOrigin
    name: SDPSessionName
    information: SDPSessionInformation | None = None
    uri: SDPSessionURI | None = None
    email: SDPSessionEmail | None = None
    phone: SDPSessionPhone | None = None
    connection: SDPSessionConnection | None = None
    bandwidth: SDPSessionBandwidth | None = None
    time: list[SDPTime] = dataclass_field(default_factory=list)
    timezone: SDPSessionTimezone | None = None
    encryption: SDPSessionEncryption | None = None
    attributes: list[SDPSessionAttributeField] = dataclass_field(default_factory=list)
    media: list[SDPMedia] = dataclass_field(default_factory=list)

    def __post_init__(self) -> None:
        if not len(self.time):
            raise ValueError("SDP session must have at least one time field")

    @property
    def mimetype(self) -> str:
        """The mimetype of the SDP session data. Always ``application/sdp``."""
        return "application/sdp"

    @property
    def media_flow_type(self) -> MediaFlowType | None:
        """The media flow type of the session, if any."""
        return get_media_flow_type(self.attributes)

    @classmethod
    def _line_preprocess(cls, line: str, fields: MutableMapping[str, Any]) -> str:
        if fields.get("media"):
            raise SDPParseError(f"Session field {line} found after media field")
        return super()._line_preprocess(line, fields)

    @property
    def connection_address(self) -> tuple[str, int] | None:
        """The advertised connection address and port to be used for media streams, if any."""
        addresses: list[tuple[str, int]] = [
            (connection.address, media.media.port)
            for media in self.media
            if (connection := media.connection or self.connection)
        ]
        if not addresses:
            return None
        if len(addresses) > 1:
            # TODO: implement allowing to filter by supported media types
            _logger.warning(
                "Multiple connection addresses found in SDP session, returning first one"
            )
        return addresses[0]
